import sys
try:
	import json
except ImportError:
	import simplejson as json

from collections import OrderedDict

from typing import (
	Union,
	Dict,
	List,
	AnyStr,
	Optional,
)
from types import ModuleType

JsonEncodable = Union[Dict, List]
# OrderedDict is also subclass of Dict, issubclass(OrderedDict, Dict) == True

def dataToPrettyJson(
	data: JsonEncodable,
	ensure_ascii: bool = False,
	sort_keys: bool = False,
):
	return json.dumps(
		data,
		sort_keys=sort_keys,
		indent=2,
		ensure_ascii=ensure_ascii,
	)


def dataToCompactJson(
	data: JsonEncodable,
	ensure_ascii: bool = False,
	sort_keys: bool = False,
) -> str:
	return json.dumps(
		data,
		sort_keys=sort_keys,
		separators=(',', ':'),
		ensure_ascii=ensure_ascii,
	)


def jsonToData(st: AnyStr) -> JsonEncodable:
	return json.loads(st)


def jsonToOrderedData(text: str) -> OrderedDict:
	return json.JSONDecoder(
		object_pairs_hook=OrderedDict,
	).decode(text)

###############################


def loadJsonConf(module: Union[ModuleType, str], confPath: str, decoders: Optional[Dict] = None) -> None:
	from os.path import isfile
	###
	if not isfile(confPath):
		return
	###
	try:
		text = open(confPath).read()
	except Exception as e:
		print('failed to read file "%s": %s' % (confPath, e))
		return
	###
	try:
		data = json.loads(text)
	except Exception as e:
		print('invalid json file "%s": %s' % (confPath, e))
		return
	###
	if isinstance(module, str):
		module = sys.modules[module]
	for param, value in data.items():
		if decoders:
			decoder = decoders.get(param)
			if decoder is not None:
				value = decoder(value)
		setattr(module, param, value)


def saveJsonConf(module: Union[ModuleType, str], confPath: str, params: str, encoders: Optional[Dict] = None) -> None:
	if isinstance(module, str):
		module = sys.modules[module]
	###
	data = OrderedDict()
	for param in params:
		value = getattr(module, param)
		if encoders:
			encoder = encoders.get(param)
			if encoder is not None:
				value = encoder(value)
		data[param] = value
	###
	text = dataToPrettyJson(data)
	try:
		open(confPath, 'w').write(text)
	except Exception as e:
		print('failed to save file "%s": %s' % (confPath, e))
		return


def loadModuleJsonConf(module: Union[ModuleType, str]) -> None:
	if isinstance(module, str):
		module = sys.modules[module]
	###
	decoders = getattr(module, 'confDecoders', {})
	###
	try:
		sysConfPath = module.sysConfPath
	except AttributeError:
		pass
	else:
		loadJsonConf(
			module,
			sysConfPath,
			decoders,
		)
	####
	loadJsonConf(
		module,
		module.confPath,
		decoders,
	)
	# should use module.confParams to restrict json keys? FIXME


def saveModuleJsonConf(module: Union[ModuleType, str]) -> None:
	if isinstance(module, str):
		module = sys.modules[module]
	###
	saveJsonConf(
		module,
		module.confPath,
		module.confParams,
		getattr(module, 'confEncoders', {}),
	)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# main.py
#
# Copyright © 2008-2010 Saeed Rasooli <saeed.gnu@gmail.com>    (ilius)
# This file is part of PyGlossary project, https://github.com/ilius/pyglossary
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
# If not, see <http://www.gnu.org/licenses/gpl.txt>.

import os
import sys
import argparse
import builtins
from os.path import dirname, join, realpath
from pprint import pformat
import logging

from pyglossary import core  # essential
from pyglossary import VERSION
from pyglossary.text_utils import startRed, endFormat

# the first thing to do is to set up logger.
# other modules also using logger "root", so it is essential to set it up prior
# to importing anything else; with exception to pyglossary.core which sets up
# logger class, and so should be done before actually initializing logger.
# verbosity level may be given on command line, so we have to parse arguments
# before setting up logger.
# once more:
# - import system modules like os, sys, argparse etc and pyglossary.core
# - parse args
# - set up logger
# - import submodules
# - other code

# no-progress-bar only for command line UI
# FIXME: load ui-dependent available options from ui modules
# (for example ui_cmd.available_options)
# the only problem is that it has to "import gtk" before it get the
# "ui_gtk.available_options"

# FIXME
# -v (verbose or version?)
# -r (reverse or read-options)

parser = argparse.ArgumentParser(add_help=False)

parser.add_argument(
	"-v",
	"--verbosity",
	action="store",
	dest="verbosity",
	type=int,
	choices=(0, 1, 2, 3, 4),
	required=False,
	default=3,  # FIXME
)
parser.add_argument(
	"--version",
	action="version",
	version="PyGlossary %s" % VERSION,
)
parser.add_argument(
	"-h",
	"--help",
	dest="help",
	action="store_true",
)
parser.add_argument(
	"-u",
	"--ui",
	dest="ui_type",
	default="auto",
	choices=(
		"cmd",
		"gtk",
		"tk",
		# "qt",
		"auto",
		"none",
	),
)
parser.add_argument(
	"-r",
	"--read-options",
	dest="readOptions",
	default="",
)
parser.add_argument(
	"-w",
	"--write-options",
	dest="writeOptions",
	default="",
)
parser.add_argument(
	"--read-format",
	dest="inputFormat",
)
parser.add_argument(
	"--write-format",
	dest="outputFormat",
	action="store",
)
parser.add_argument(
	"--direct",
	dest="direct",
	action="store_true",
	default=None,
	help="if possible, convert directly without loading into memory",
)
parser.add_argument(
	"--indirect",
	dest="direct",
	action="store_false",
	default=None,
	help="disable `direct` mode, load full data into memory before writing"
		 ", this is default",
)
parser.add_argument(
	"--reverse",
	dest="reverse",
	action="store_true",
)
parser.add_argument(
	"--no-progress-bar",
	dest="progressbar",
	action="store_false",
	default=None,
)
parser.add_argument(
	"--sort",
	dest="sort",
	action="store_true",
	default=None,
)
parser.add_argument(
	"--no-sort",
	dest="sort",
	action="store_false",
	default=None,
)
parser.add_argument(
	"--sort-cache-size",
	dest="sortCacheSize",
	type=int,
	default=None,
)

parser.add_argument(
	"--skip-resources",
	dest="skipResources",
	action="store_true",
	default=None,
	help="skip resources (images, audio, etc)",
)
parser.add_argument(
	"--utf8-check",
	dest="utf8Check",
	action="store_true",
	default=None,
)
parser.add_argument(
	"--no-utf8-check",
	dest="utf8Check",
	action="store_false",
	default=None,
)
parser.add_argument(
	"--lower",
	dest="lower",
	action="store_true",
	default=None,
	help="lowercase words before writing",
)
parser.add_argument(
	"--no-lower",
	dest="lower",
	action="store_false",
	default=None,
	help="do not lowercase words before writing",
)
parser.add_argument(
	"--remove-html",
	dest="remove_html",
	help="remove given html tags (comma-separated) from definitions",
)
parser.add_argument(
	"--remove-html-all",
	dest="remove_html_all",
	action="store_true",
	help="remove all html tags from definitions",
)


parser.add_argument(
	"--no-color",
	dest="noColor",
	action="store_true",
)
parser.add_argument(
	"inputFilename",
	action="store",
	default="",
	nargs="?",
)
parser.add_argument(
	"outputFilename",
	action="store",
	default="",
	nargs="?",
)

args = parser.parse_args()

log = logging.getLogger("root")

defaultVerbosity = log.getVerbosity()

log.setVerbosity(args.verbosity)
log.addHandler(
	core.StdLogHandler(noColor=args.noColor),
)
# with the logger setted up, we can import other pyglossary modules, so they
# can do some loggging in right way.


core.checkCreateConfDir()


##############################

from pyglossary.glossary import Glossary
from ui.ui_cmd import COMMAND, help, parseFormatOptionsStr

if args.verbosity != defaultVerbosity:
	Glossary.init()

##############################


def dashToCamelCase(text):  # converts "hello-PYTHON-user" to "helloPythonUser"
	parts = text.split("-")
	parts[0] = parts[0].lower()
	for i in range(1, len(parts)):
		parts[i] = parts[i].capitalize()
	return "".join(parts)

ui_list = (
	"gtk",
	"tk",
	"qt",
)

# log.info("PyGlossary %s"%VERSION)


if args.help:
	help()
	sys.exit(0)


if os.sep != "/":
	args.noColor = True


# only used in ui_cmd for now
readOptions = parseFormatOptionsStr(args.readOptions)
writeOptions = parseFormatOptionsStr(args.writeOptions)


"""
	examples for read and write options:
	--read-options testOption=stringValue
	--read-options enableFoo=True
	--read-options fooList=[1,2,3]
	--read-options 'fooList=[1, 2, 3]'
	--read-options 'testOption=stringValue; enableFoo=True; fooList=[1, 2, 3]'
	--read-options 'testOption=stringValue;enableFoo=True;fooList=[1,2,3]'
"""

# FIXME
prefOptionsKeys = (
	# "verbosity",
	"utf8Check",
	"lower",
	"skipResources",
	"remove_html",
	"remove_html_all",
)

convertOptionsKeys = (
	"direct",
	"progressbar",
	"sort",
	"sortCacheSize",
	# "sortKey",# or sortAlg FIXME
)

prefOptions = {}
for param in prefOptionsKeys:
	value = getattr(args, param, None)
	if value is not None:
		prefOptions[param] = value

convertOptions = {}
for param in convertOptionsKeys:
	value = getattr(args, param, None)
	if value is not None:
		convertOptions[param] = value


if args.inputFilename and readOptions:
	inputFormat = Glossary.detectInputFormat(args.inputFilename, format=args.inputFormat)
	if not inputFormat:
		log.error("Could not detect format for input file %s" % args.inputFilename)
		sys.exit(1)
	readOptionsProp = Glossary.formatsOptionsProp[inputFormat]
	for optName, optValue in readOptions.items():
		if optName not in Glossary.formatsReadOptions[inputFormat]:
			log.error("Invalid option name %s for format %s" % (optName, inputFormat))
			sys.exit(1)
		prop = readOptionsProp[optName]
		optValueNew, ok = prop.evaluate(optValue)
		if not ok or not prop.validate(optValueNew):
			log.error("Invalid option value %s=%r for format %s" % (optName, optValue, inputFormat))
			sys.exit(1)
		readOptions[optName] = optValueNew

if args.outputFilename and writeOptions:
	_, outputFormat, _ = Glossary.detectOutputFormat(
		filename=args.outputFilename,
		format=args.outputFormat,
		inputFilename=args.inputFilename,
	)
	if not outputFormat:
		log.error("Could not detect format for output file %s" % args.outputFilename)
		sys.exit(1)
	writeOptionsProp = Glossary.formatsOptionsProp[outputFormat]
	for optName, optValue in writeOptions.items():
		if optName not in Glossary.formatsWriteOptions[outputFormat]:
			log.error("Invalid option name %s for format %s" % (optName, outputFormat))
			sys.exit(1)
		prop = writeOptionsProp[optName]
		optValueNew, ok = prop.evaluate(optValue)
		if not ok or not prop.validate(optValueNew):
			log.error("Invalid option value %s=%r for format %s" % (optName, optValue, outputFormat))
			sys.exit(1)
		writeOptions[optName] = optValueNew


if prefOptions:
	log.debug("prefOptions = %s", prefOptions)
if convertOptions:
	log.debug("convertOptions = %s", convertOptions)


"""
ui_type: User interface type
Possible values:
	cmd - Command line interface, this ui will automatically selected
		  if you give both input and output file
	gtk - GTK interface
	tk - Tkinter interface
	qt - Qt interface
	auto - Use the first available UI
"""
ui_type = args.ui_type


if args.inputFilename:
	if args.outputFilename and ui_type != "none":
		ui_type = "cmd"  # silently? FIXME
else:
	if ui_type == "cmd":
		log.error("no input file given, try --help")
		exit(1)


if ui_type == "none":
	if args.reverse:
		log.error("--reverse does not work with --ui=none")
		sys.exit(1)
	glos = Glossary()
	glos.convert(
		args.inputFilename,
		inputFormat=args.inputFormat,
		outputFilename=args.outputFilename,
		outputFormat=args.outputFormat,
		readOptions=readOptions,
		writeOptions=writeOptions,
		**convertOptions
	)
	sys.exit(0)
elif ui_type == "cmd":
	from ui import ui_cmd
	sys.exit(0 if ui_cmd.UI().run(
		args.inputFilename,
		outputFilename=args.outputFilename,
		inputFormat=args.inputFormat,
		outputFormat=args.outputFormat,
		reverse=args.reverse,
		prefOptions=prefOptions,
		readOptions=readOptions,
		writeOptions=writeOptions,
		convertOptions=convertOptions,
	) else 1)
if ui_type == "auto":
	ui_module = None
	for ui_type2 in ui_list:
		try:
			ui_module = getattr(
				__import__("ui.ui_%s" % ui_type2),
				"ui_%s" % ui_type2,
			)
		except ImportError:
			log.exception("error while importing UI module:")  # FIXME
		else:
			break
	if ui_module is None:
		log.error("no user interface module found! try \"%s -h\" to see command line usage" % sys.argv[0])
		sys.exit(1)
else:
	ui_module = getattr(
		__import__("ui.ui_%s" % ui_type),
		"ui_%s" % ui_type,
	)

sys.exit(0 if ui_module.UI(**prefOptions).run(
	editPath=args.inputFilename,
	readOptions=readOptions,
) else 1)



"""Functions that read and write gzipped files.

The user of the file doesn't have to worry about the compression,
but random access is not allowed."""

# based on Andrew Kuchling's minigzip.py distributed with the zlib module

import struct, sys, time, os
import zlib
import builtins
import io

__all__ = ["GzipFile", "open", "compress", "decompress"]

FTEXT, FHCRC, FEXTRA, FNAME, FCOMMENT = 1, 2, 4, 8, 16

READ, WRITE = 1, 2

def open(filename, mode="rb", compresslevel=9,
         encoding=None, errors=None, newline=None):
    """Open a gzip-compressed file in binary or text mode.

    The filename argument can be an actual filename (a str or bytes object), or
    an existing file object to read from or write to.

    The mode argument can be "r", "rb", "w", "wb", "x", "xb", "a" or "ab" for
    binary mode, or "rt", "wt", "xt" or "at" for text mode. The default mode is
    "rb", and the default compresslevel is 9.

    For binary mode, this function is equivalent to the GzipFile constructor:
    GzipFile(filename, mode, compresslevel). In this case, the encoding, errors
    and newline arguments must not be provided.

    For text mode, a GzipFile object is created, and wrapped in an
    io.TextIOWrapper instance with the specified encoding, error handling
    behavior, and line ending(s).

    """
    if "t" in mode:
        if "b" in mode:
            raise ValueError("Invalid mode: %r" % (mode,))
    else:
        if encoding is not None:
            raise ValueError("Argument 'encoding' not supported in binary mode")
        if errors is not None:
            raise ValueError("Argument 'errors' not supported in binary mode")
        if newline is not None:
            raise ValueError("Argument 'newline' not supported in binary mode")

    gz_mode = mode.replace("t", "")
    if isinstance(filename, (str, bytes)):
        binary_file = GzipFile(filename, gz_mode, compresslevel)
    elif hasattr(filename, "read") or hasattr(filename, "write"):
        binary_file = GzipFile(None, gz_mode, compresslevel, filename)
    else:
        raise TypeError("filename must be a str or bytes object, or a file")

    if "t" in mode:
        return io.TextIOWrapper(binary_file, encoding, errors, newline)
    else:
        return binary_file

def write32u(output, value):
    # The L format writes the bit pattern correctly whether signed
    # or unsigned.
    output.write(struct.pack("<L", value))

class _PaddedFile:
    """Minimal read-only file object that prepends a string to the contents
    of an actual file. Shouldn't be used outside of gzip.py, as it lacks
    essential functionality."""

    def __init__(self, f, prepend=b''):
        self._buffer = prepend
        self._length = len(prepend)
        self.file = f
        self._read = 0

    def read(self, size):
        if self._read is None:
            return self.file.read(size)
        if self._read + size <= self._length:
            read = self._read
            self._read += size
            return self._buffer[read:self._read]
        else:
            read = self._read
            self._read = None
            return self._buffer[read:] + \
                   self.file.read(size-self._length+read)

    def prepend(self, prepend=b'', readprevious=False):
        if self._read is None:
            self._buffer = prepend
        elif readprevious and len(prepend) <= self._read:
            self._read -= len(prepend)
            return
        else:
            self._buffer = self._buffer[self._read:] + prepend
        self._length = len(self._buffer)
        self._read = 0

    def unused(self):
        if self._read is None:
            return b''
        return self._buffer[self._read:]

    def seek(self, offset, whence=0):
        # This is only ever called with offset=whence=0
        if whence == 1 and self._read is not None:
            if 0 <= offset + self._read <= self._length:
                self._read += offset
                return
            else:
                offset += self._length - self._read
        self._read = None
        self._buffer = None
        return self.file.seek(offset, whence)

    def __getattr__(self, name):
        return getattr(self.file, name)


class GzipFile(io.BufferedIOBase):
    """The GzipFile class simulates most of the methods of a file object with
    the exception of the readinto() and truncate() methods.

    This class only supports opening files in binary mode. If you need to open a
    compressed file in text mode, use the gzip.open() function.

    """

    myfileobj = None
    max_read_chunk = 10 * 1024 * 1024   # 10Mb

    def __init__(self, filename=None, mode=None,
                 compresslevel=9, fileobj=None, mtime=None):
        """Constructor for the GzipFile class.

        At least one of fileobj and filename must be given a
        non-trivial value.

        The new class instance is based on fileobj, which can be a regular
        file, an io.BytesIO object, or any other object which simulates a file.
        It defaults to None, in which case filename is opened to provide
        a file object.

        When fileobj is not None, the filename argument is only used to be
        included in the gzip file header, which may includes the original
        filename of the uncompressed file.  It defaults to the filename of
        fileobj, if discernible; otherwise, it defaults to the empty string,
        and in this case the original filename is not included in the header.

        The mode argument can be any of 'r', 'rb', 'a', 'ab', 'w', 'wb', 'x', or
        'xb' depending on whether the file will be read or written.  The default
        is the mode of fileobj if discernible; otherwise, the default is 'rb'.
        A mode of 'r' is equivalent to one of 'rb', and similarly for 'w' and
        'wb', 'a' and 'ab', and 'x' and 'xb'.

        The compresslevel argument is an integer from 0 to 9 controlling the
        level of compression; 1 is fastest and produces the least compression,
        and 9 is slowest and produces the most compression. 0 is no compression
        at all. The default is 9.

        The mtime argument is an optional numeric timestamp to be written
        to the stream when compressing.  All gzip compressed streams
        are required to contain a timestamp.  If omitted or None, the
        current time is used.  This module ignores the timestamp when
        decompressing; however, some programs, such as gunzip, make use
        of it.  The format of the timestamp is the same as that of the
        return value of time.time() and of the st_mtime member of the
        object returned by os.stat().

        """

        if mode and ('t' in mode or 'U' in mode):
            raise ValueError("Invalid mode: {!r}".format(mode))
        if mode and 'b' not in mode:
            mode += 'b'
        if fileobj is None:
            fileobj = self.myfileobj = builtins.open(filename, mode or 'rb')
        if filename is None:
            filename = getattr(fileobj, 'name', '')
            if not isinstance(filename, (str, bytes)):
                filename = ''
        if mode is None:
            mode = getattr(fileobj, 'mode', 'rb')

        if mode.startswith('r'):
            self.mode = READ
            # Set flag indicating start of a new member
            self._new_member = True
            # Buffer data read from gzip file. extrastart is offset in
            # stream where buffer starts. extrasize is number of
            # bytes remaining in buffer from current stream position.
            self.extrabuf = b""
            self.extrasize = 0
            self.extrastart = 0
            self.name = filename
            # Starts small, scales exponentially
            self.min_readsize = 100
            fileobj = _PaddedFile(fileobj)

        elif mode.startswith(('w', 'a', 'x')):
            self.mode = WRITE
            self._init_write(filename)
            self.compress = zlib.compressobj(compresslevel,
                                             zlib.DEFLATED,
                                             -zlib.MAX_WBITS,
                                             zlib.DEF_MEM_LEVEL,
                                             0)
        else:
            raise ValueError("Invalid mode: {!r}".format(mode))

        self.fileobj = fileobj
        self.offset = 0
        self.mtime = mtime

        if self.mode == WRITE:
            self._write_gzip_header()

    @property
    def filename(self):
        import warnings
        warnings.warn("use the name attribute", DeprecationWarning, 2)
        if self.mode == WRITE and self.name[-3:] != ".gz":
            return self.name + ".gz"
        return self.name

    def __repr__(self):
        fileobj = self.fileobj
        if isinstance(fileobj, _PaddedFile):
            fileobj = fileobj.file
        s = repr(fileobj)
        return '<gzip ' + s[1:-1] + ' ' + hex(id(self)) + '>'

    def _check_closed(self):
        """Raises a ValueError if the underlying file object has been closed.

        """
        if self.closed:
            raise ValueError('I/O operation on closed file.')

    def _init_write(self, filename):
        self.name = filename
        self.crc = zlib.crc32(b"") & 0xffffffff
        self.size = 0
        self.writebuf = []
        self.bufsize = 0

    def _write_gzip_header(self):
        self.fileobj.write(b'\037\213')             # magic header
        self.fileobj.write(b'\010')                 # compression method
        try:
            # RFC 1952 requires the FNAME field to be Latin-1. Do not
            # include filenames that cannot be represented that way.
            fname = os.path.basename(self.name)
            if not isinstance(fname, bytes):
                fname = fname.encode('latin-1')
            if fname.endswith(b'.gz'):
                fname = fname[:-3]
        except UnicodeEncodeError:
            fname = b''
        flags = 0
        if fname:
            flags = FNAME
        self.fileobj.write(chr(flags).encode('latin-1'))
        mtime = self.mtime
        if mtime is None:
            mtime = time.time()
        write32u(self.fileobj, int(mtime))
        self.fileobj.write(b'\002')
        self.fileobj.write(b'\377')
        if fname:
            self.fileobj.write(fname + b'\000')

    def _init_read(self):
        self.crc = zlib.crc32(b"") & 0xffffffff
        self.size = 0

    def _read_exact(self, n):
        data = self.fileobj.read(n)
        while len(data) < n:
            b = self.fileobj.read(n - len(data))
            if not b:
                raise EOFError("Compressed file ended before the "
                               "end-of-stream marker was reached")
            data += b
        return data

    def _read_gzip_header(self):
        magic = self.fileobj.read(2)
        if magic == b'':
            return False

        if magic != b'\037\213':
            raise OSError('Not a gzipped file')

        method, flag, self.mtime = struct.unpack("<BBIxx", self._read_exact(8))
        if method != 8:
            raise OSError('Unknown compression method')

        if flag & FEXTRA:
            # Read & discard the extra field, if present
            extra_len, = struct.unpack("<H", self._read_exact(2))
            self._read_exact(extra_len)
        if flag & FNAME:
            # Read and discard a null-terminated string containing the filename
            while True:
                s = self.fileobj.read(1)
                if not s or s==b'\000':
                    break
        if flag & FCOMMENT:
            # Read and discard a null-terminated string containing a comment
            while True:
                s = self.fileobj.read(1)
                if not s or s==b'\000':
                    break
        if flag & FHCRC:
            self._read_exact(2)     # Read & discard the 16-bit header CRC

        unused = self.fileobj.unused()
        if unused:
            uncompress = self.decompress.decompress(unused)
            self._add_read_data(uncompress)
        return True

    def write(self,data):
        self._check_closed()
        if self.mode != WRITE:
            import errno
            raise OSError(errno.EBADF, "write() on read-only GzipFile object")

        if self.fileobj is None:
            raise ValueError("write() on closed GzipFile object")

        # Convert data type if called by io.BufferedWriter.
        if isinstance(data, memoryview):
            data = data.tobytes()

        if len(data) > 0:
            self.fileobj.write(self.compress.compress(data))
            self.size += len(data)
            self.crc = zlib.crc32(data, self.crc) & 0xffffffff
            self.offset += len(data)

        return len(data)

    def read(self, size=-1):
        self._check_closed()
        if self.mode != READ:
            import errno
            raise OSError(errno.EBADF, "read() on write-only GzipFile object")

        if self.extrasize <= 0 and self.fileobj is None:
            return b''

        readsize = 1024
        if size < 0:        # get the whole thing
            while self._read(readsize):
                readsize = min(self.max_read_chunk, readsize * 2)
            size = self.extrasize
        else:               # just get some more of it
            while size > self.extrasize:
                if not self._read(readsize):
                    if size > self.extrasize:
                        size = self.extrasize
                    break
                readsize = min(self.max_read_chunk, readsize * 2)

        offset = self.offset - self.extrastart
        chunk = self.extrabuf[offset: offset + size]
        self.extrasize = self.extrasize - size

        self.offset += size
        return chunk

    def read1(self, size=-1):
        self._check_closed()
        if self.mode != READ:
            import errno
            raise OSError(errno.EBADF, "read1() on write-only GzipFile object")

        if self.extrasize <= 0 and self.fileobj is None:
            return b''

        # For certain input data, a single call to _read() may not return
        # any data. In this case, retry until we get some data or reach EOF.
        while self.extrasize <= 0 and self._read():
            pass
        if size < 0 or size > self.extrasize:
            size = self.extrasize

        offset = self.offset - self.extrastart
        chunk = self.extrabuf[offset: offset + size]
        self.extrasize -= size
        self.offset += size
        return chunk

    def peek(self, n):
        if self.mode != READ:
            import errno
            raise OSError(errno.EBADF, "peek() on write-only GzipFile object")

        # Do not return ridiculously small buffers, for one common idiom
        # is to call peek(1) and expect more bytes in return.
        if n < 100:
            n = 100
        if self.extrasize == 0:
            if self.fileobj is None:
                return b''
            # Ensure that we don't return b"" if we haven't reached EOF.
            # 1024 is the same buffering heuristic used in read()
            while self.extrasize == 0 and self._read(max(n, 1024)):
                pass
        offset = self.offset - self.extrastart
        remaining = self.extrasize
        assert remaining == len(self.extrabuf) - offset
        return self.extrabuf[offset:offset + n]

    def _unread(self, buf):
        self.extrasize = len(buf) + self.extrasize
        self.offset -= len(buf)

    def _read(self, size=1024):
        if self.fileobj is None:
            return False

        if self._new_member:
            # If the _new_member flag is set, we have to
            # jump to the next member, if there is one.
            self._init_read()
            if not self._read_gzip_header():
                return False
            self.decompress = zlib.decompressobj(-zlib.MAX_WBITS)
            self._new_member = False

        # Read a chunk of data from the file
        buf = self.fileobj.read(size)

        # If the EOF has been reached, flush the decompression object
        # and mark this object as finished.

        if buf == b"":
            uncompress = self.decompress.flush()
            # Prepend the already read bytes to the fileobj to they can be
            # seen by _read_eof()
            self.fileobj.prepend(self.decompress.unused_data, True)
            self._read_eof()
            self._add_read_data( uncompress )
            return False

        uncompress = self.decompress.decompress(buf)
        self._add_read_data( uncompress )

        if self.decompress.unused_data != b"":
            # Ending case: we've come to the end of a member in the file,
            # so seek back to the start of the unused data, finish up
            # this member, and read a new gzip header.
            # Prepend the already read bytes to the fileobj to they can be
            # seen by _read_eof() and _read_gzip_header()
            self.fileobj.prepend(self.decompress.unused_data, True)
            # Check the CRC and file size, and set the flag so we read
            # a new member on the next call
            self._read_eof()
            self._new_member = True
        return True

    def _add_read_data(self, data):
        self.crc = zlib.crc32(data, self.crc) & 0xffffffff
        offset = self.offset - self.extrastart
        self.extrabuf = self.extrabuf[offset:] + data
        self.extrasize = self.extrasize + len(data)
        self.extrastart = self.offset
        self.size = self.size + len(data)

    def _read_eof(self):
        # We've read to the end of the file
        # We check the that the computed CRC and size of the
        # uncompressed data matches the stored values.  Note that the size
        # stored is the true file size mod 2**32.
        crc32, isize = struct.unpack("<II", self._read_exact(8))
        if crc32 != self.crc:
            raise OSError("CRC check failed %s != %s" % (hex(crc32),
                                                         hex(self.crc)))
        elif isize != (self.size & 0xffffffff):
            raise OSError("Incorrect length of data produced")

        # Gzip files can be padded with zeroes and still have archives.
        # Consume all zero bytes and set the file position to the first
        # non-zero byte. See http://www.gzip.org/#faq8
        c = b"\x00"
        while c == b"\x00":
            c = self.fileobj.read(1)
        if c:
            self.fileobj.prepend(c, True)

    @property
    def closed(self):
        return self.fileobj is None

    def close(self):
        fileobj = self.fileobj
        if fileobj is None:
            return
        self.fileobj = None
        try:
            if self.mode == WRITE:
                fileobj.write(self.compress.flush())
                write32u(fileobj, self.crc)
                # self.size may exceed 2GB, or even 4GB
                write32u(fileobj, self.size & 0xffffffff)
        finally:
            myfileobj = self.myfileobj
            if myfileobj:
                self.myfileobj = None
                myfileobj.close()

    def flush(self,zlib_mode=zlib.Z_SYNC_FLUSH):
        self._check_closed()
        if self.mode == WRITE:
            # Ensure the compressor's buffer is flushed
            self.fileobj.write(self.compress.flush(zlib_mode))
            self.fileobj.flush()

    def fileno(self):
        """Invoke the underlying file object's fileno() method.

        This will raise AttributeError if the underlying file object
        doesn't support fileno().
        """
        return self.fileobj.fileno()

    def rewind(self):
        '''Return the uncompressed stream file position indicator to the
        beginning of the file'''
        if self.mode != READ:
            raise OSError("Can't rewind in write mode")
        self.fileobj.seek(0)
        self._new_member = True
        self.extrabuf = b""
        self.extrasize = 0
        self.extrastart = 0
        self.offset = 0

    def readable(self):
        return self.mode == READ

    def writable(self):
        return self.mode == WRITE

    def seekable(self):
        return True

    def seek(self, offset, whence=0):
        if whence:
            if whence == 1:
                offset = self.offset + offset
            else:
                raise ValueError('Seek from end not supported')
        if self.mode == WRITE:
            if offset < self.offset:
                raise OSError('Negative seek in write mode')
            count = offset - self.offset
            chunk = bytes(1024)
            for i in range(count // 1024):
                self.write(chunk)
            self.write(bytes(count % 1024))
        elif self.mode == READ:
            if offset < self.offset:
                # for negative seek, rewind and do positive seek
                self.rewind()
            count = offset - self.offset
            for i in range(count // 1024):
                self.read(1024)
            self.read(count % 1024)

        return self.offset

    def readline(self, size=-1):
        if size < 0:
            # Shortcut common case - newline found in buffer.
            offset = self.offset - self.extrastart
            i = self.extrabuf.find(b'\n', offset) + 1
            if i > 0:
                self.extrasize -= i - offset
                self.offset += i - offset
                return self.extrabuf[offset: i]

            size = sys.maxsize
            readsize = self.min_readsize
        else:
            readsize = size
        bufs = []
        while size != 0:
            c = self.read(readsize)
            i = c.find(b'\n')

            # We set i=size to break out of the loop under two
            # conditions: 1) there's no newline, and the chunk is
            # larger than size, or 2) there is a newline, but the
            # resulting line would be longer than 'size'.
            if (size <= i) or (i == -1 and len(c) > size):
                i = size - 1

            if i >= 0 or c == b'':
                bufs.append(c[:i + 1])    # Add portion of last chunk
                self._unread(c[i + 1:])   # Push back rest of chunk
                break

            # Append chunk to list, decrease 'size',
            bufs.append(c)
            size = size - len(c)
            readsize = min(size, readsize * 2)
        if readsize > self.min_readsize:
            self.min_readsize = min(readsize, self.min_readsize * 2, 512)
        return b''.join(bufs) # Return resulting line


def compress(data, compresslevel=9):
    """Compress data in one shot and return the compressed string.
    Optional argument is the compression level, in range of 0-9.
    """
    buf = io.BytesIO()
    with GzipFile(fileobj=buf, mode='wb', compresslevel=compresslevel) as f:
        f.write(data)
    return buf.getvalue()

def decompress(data):
    """Decompress a gzip compressed string in one shot.
    Return the decompressed string.
    """
    with GzipFile(fileobj=io.BytesIO(data)) as f:
        return f.read()


def _test():
    # Act like gzip; with -d, act like gunzip.
    # The input file is not deleted, however, nor are any other gzip
    # options or features supported.
    args = sys.argv[1:]
    decompress = args and args[0] == "-d"
    if decompress:
        args = args[1:]
    if not args:
        args = ["-"]
    for arg in args:
        if decompress:
            if arg == "-":
                f = GzipFile(filename="", mode="rb", fileobj=sys.stdin.buffer)
                g = sys.stdout.buffer
            else:
                if arg[-3:] != ".gz":
                    print("filename doesn't end in .gz:", repr(arg))
                    continue
                f = open(arg, "rb")
                g = builtins.open(arg[:-3], "wb")
        else:
            if arg == "-":
                f = sys.stdin.buffer
                g = GzipFile(filename="", mode="wb", fileobj=sys.stdout.buffer)
            else:
                f = builtins.open(arg, "rb")
                g = open(arg + ".gz", "wb")
        while True:
            chunk = f.read(1024)
            if not chunk:
                break
            g.write(chunk)
        if g is not sys.stdout.buffer:
            g.close()
        if f is not sys.stdin.buffer:
            f.close()

if __name__ == '__main__':
    _test()



# -*- coding: utf-8 -*-
# The MIT License (MIT)

# Copyright (C) 2012-2016 Alberto Pettarin (alberto@albertopettarin.it)
# Copyright (C) 2016 Saeed Rasooli <saeed.gnu@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from formats_common import *
from pyglossary.ebook_base import get_prefix
from itertools import groupby
from pathlib import Path

enable = True
format = "Kobo"
description = "Kobo E-Reader Dictionary"
extensions = [".kobo"]
sortOnWrite = ALWAYS

optionsProp = {
	"group_by_prefix_length": IntOption(),
	# "group_by_prefix_merge_min_size": IntOption(),
	# "group_by_prefix_merge_across_first": BoolOption(),
	"keep": BoolOption(),
	"marisa_bin_path": StrOption(),
}


def fixFilename(fname: str) -> str:
	return Path(fname.replace("/", "2F").replace("\\", "5C")).name


class Writer:
	WORDS_FILE_NAME = "words"
	MARISA_BUILD = "marisa-build"
	MARISA_REVERSE_LOOKUP = "marisa-reverse-lookup"

	def __init__(self, glos, **kwargs):
		self._glos = glos

	def write_groups(self, group_by_prefix_length):
		import gzip
		words = []

		self._glos.sortWords()
		for group_i, (group_prefix, group_entry_iter) in enumerate(groupby(
			self._glos,
			lambda tmpEntry: get_prefix(
				tmpEntry.getWord(),
				group_by_prefix_length,
			),
		)):
			group_fname = fixFilename(group_prefix)
			htmlContents = "<?xml version=\"1.0\" encoding=\"utf-8\"?><html>\n"
			for entry in group_entry_iter:
				if entry.isData():
					continue
				word = entry.getWord()
				defi = entry.getDefi()
				words.append(word)
				htmlContents += f"<w><a name=\"{word}\"/><div><b>{word}</b>"\
					f"<br/>{defi}</div></w>\n"
			htmlContents += "</html>"
			with gzip.open(group_fname + ".html", mode="wb") as gzipFile:
				gzipFile.write(htmlContents.encode("utf-8"))

		return words

	def write(
		self,
		filename,
		group_by_prefix_length=2,
	):
		try:
			import marisa_trie
		except ImportError as e:
			log.error("Run: sudo pip3 install marisa-trie")
			raise e

		with indir(filename, create=True):
			words = self.write_groups(group_by_prefix_length)
			trie = marisa_trie.Trie(words)
			trie.save(self.WORDS_FILE_NAME)



import re
import os
from .pinyin import convert
from .summarize import summarize

line_reg = re.compile(r"^([^ ]+) ([^ ]+) \[([^\]]+)\] /(.+)/$")

jinja_env = None

script_dir = os.path.dirname(__file__)

COLORS = {
	"": "black",
	"1": "red",
	"2": "orange",
	"3": "green",
	"4": "blue",
	"5": "black",
}

try:
	ModuleNotFoundError
except NameError:
	ModuleNotFoundError = ImportError

def load_jinja():
	global jinja_env
	try:
		import jinja2
	except ModuleNotFoundError as e:
		e.msg += ", run `sudo pip3 install jinja2` to install"
		raise e
	from .jinja2htmlcompress import HTMLCompress
	jinja_env = jinja2.Environment(
		loader=jinja2.FileSystemLoader(script_dir),
		extensions=[HTMLCompress],
	)

def parse_line(line):
	line = line.strip()
	match = line_reg.match(line)
	if match is None: return None
	trad, simp, pinyin, eng = match.groups()
	pinyin = pinyin.replace("u:", "v")
	eng = eng.split("/")
	return trad, simp, pinyin, eng

def make_entry(trad, simp, pinyin, eng):
	eng_names = list(map(summarize, eng))
	names = [simp, trad, pinyin] + eng_names
	article = render_article(trad, simp, pinyin, eng)
	return names, article

def render_article(trad, simp, pinyin, eng):
	if jinja_env is None:
		load_jinja()

	pinyin_tones = [convert(syl) for syl in pinyin.split()]
	nice_pinyin = []
	tones = []
	for syllable in pinyin.split():
		nice_syllable, tone = convert(syllable)
		nice_pinyin.append(nice_syllable)
		tones.append(tone)

	template = jinja_env.get_template("article.html")
	return template.render(
		zip=zip,
		COLORS=COLORS,
		trad=trad,
		simp=simp,
		pinyin=nice_pinyin,
		tones=tones,
		defns=eng,
	)




from gi.repository import Gtk as gtk
from gi.repository import Gdk as gdk


class MyDialog(object):
	def startWaiting(self):
		self.queue_draw()
		self.vbox.set_sensitive(False)
		self.get_window().set_cursor(gdk.Cursor.new(gdk.CursorType.WATCH))
		while gtk.events_pending():
			gtk.main_iteration_do(False)

	def endWaiting(self):
		self.get_window().set_cursor(gdk.Cursor.new(gdk.CursorType.LEFT_PTR))
		self.vbox.set_sensitive(True)

	def waitingDo(self, func, *args, **kwargs):
		self.startWaiting()
		try:
			func(*args, **kwargs)
		except Exception as e:
			raise e
		finally:
			self.endWaiting()



from itertools import (
	takewhile,
	repeat,
)

from typing import (
	AnyStr,
	Iterator
)
from io import IOBase


def toBytes(s: AnyStr) -> bytes:
	return bytes(s, "utf-8") if isinstance(s, str) else bytes(s)


def fileCountLines(filename: str, newline: str = "\n"):
	newline = toBytes(newline)  # required? FIXME
	f = open(filename, "rb")  # or "r"
	bufgen = takewhile(
		lambda x: x, (f.read(1024*1024) for _ in repeat(None))
	)
	return sum(
		buf.count(newline) for buf in bufgen if buf
	)


# TODO: make it sub-class of IOBase
class FileLineWrapper(object):
	def __init__(self, f: IOBase):
		self.f = f
		self.line = 0

	def close(self) -> None:
		self.f.close()

	def readline(self) -> str:
		self.line += 1
		return self.f.readline()

	def __iter__(self) -> Iterator[str]:
		return iter(self.f)



from . import *
from .utils import *


class ResizeButton(gtk.EventBox):
	def __init__(self, win, edge=gdk.WindowEdge.SOUTH_EAST):
		gtk.EventBox.__init__(self)
		self.win = win
		self.edge = edge
		###
		self.image = imageFromFile('resize.png')
		self.add(self.image)
		self.connect('button-press-event', self.buttonPress)

	def buttonPress(self, obj, gevent):
		self.win.begin_resize_drag(
			self.edge,
			gevent.button,
			int(gevent.x_root),
			int(gevent.y_root),
			gevent.time,
		)



# -*- coding: utf-8 -*-

from formats_common import *
from pyglossary.entry import Entry

enable = True
format = 'Aard2Slob'
description = 'Aard 2 (slob)'
extensions = ['.slob']
optionsProp = {
	"compression": StrOption(
		values=["", "bz2", "zlib", "lzma2"],
		comment="Compression Algorithm",
	),
	"content_type": StrOption(
		customValue=True,
		values=["text/plain; charset=utf-8"],
		comment="Content Type",
	),
	"encoding": EncodingOption(),
}
depends = {
	"icu": "PyICU",
}


class Reader(object):
	def __init__(self, glos):
		self._glos = glos
		self._clear()

	def close(self):
		if self._slobObj is not None:
			self._slobObj.close()
		self._clear()

	def _clear(self):
		self._filename = ""
		self._slobObj = None # slobObj is instance of slob.Slob class
		self._refIndex = -1

	def open(self, filename, encoding="utf-8"):
		from pyglossary.plugin_lib import slob
		self._filename = filename
		self._slobObj = slob.open(filename)

	def __len__(self):
		if self._slobObj is None:
			log.error("called len() on a reader which is not open")
			return 0
		return len(self._slobObj)

	def __iter__(self):
		return self

	def __next__(self):
		if not self._slobObj:
			log.error("iterating over a reader which is not open")
			raise StopIteration
		self._refIndex += 1
		if self._refIndex >= len(self._slobObj):
			raise StopIteration
		blob = self._slobObj[self._refIndex]
		# blob.key is str, blob.content is bytes
		word = blob.key
		defi = toStr(blob.content)
		return Entry(word, defi)


def write(
	glos: GlossaryType,
	filename: str,
	compression: str = "",
	content_type: str = "",
):
	from pyglossary.plugin_lib import slob
	kwargs = {}
	if compression:
		kwargs["compression"] = compression
	# must not pass compression=None to slob.create()
	with slob.create(filename, **kwargs) as slobWriter:
		name = glos.getInfo("name")
		slobWriter.tag("label", toStr(name))
		for entry in glos:
			words = entry.getWords()
			b_defi = entry.getDefi().encode("utf-8")
			slobWriter.add(
				b_defi,
				*tuple(words),
				content_type=content_type,
			)
	# slobWriter.finalize() is called called on __exit__


# is Writer class needed?






# -*- coding: utf-8 -*-

# Copyright (C) 2019 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

from formats_common import *

from struct import unpack
from zlib import decompress
from lxml import etree

enable = True
format = "AppleDictBin"
description = "AppleDict Binary (.dictionary)"
extensions = [".dictionary", ".data"]
optionsProp = {
	"html": BoolOption(),
}


class Reader(object):
	def __init__(self, glos):
		self._glos = glos
		self._filename = ""
		self._file = None
		self._encoding = "utf-8"
		self._buf = ""
		self._html = False
		self._defiFormat = "m"

	def open(self, filename, html: bool = False):
		self._html = html
		self._defiFormat = "h" if html else "m"
		parts = filename.split(os.sep)
		dbname = parts[-1]
		if isdir(filename):
			if parts[-1] == "Contents":
				filename = join(filename, "Body.data")
				if len(parts) > 2:
					dbname = parts[-2]
			else:
				filename = join(filename, "Contents/Resources/Body.data")
		elif dbname == "Body.data" and len(parts) > 1:
			dbname = parts[-2]
			if len(parts) > 2:
				if dbname == "Contents":
					dbname = parts[-3]
				elif dbname == "Resources" and len(parts) > 3:
					dbname = parts[-4]

		if dbname.endswith(".dictionary"):
			dbname = dbname[:-len(".dictionary")]
		self._glos.setInfo("name", dbname)

		self._filename = filename
		self._file = open(filename, "rb")

		self._file.seek(0x40)
		self._limit = 0x40 + unpack("i", self._file.read(4))[0]
		self._file.seek(0x60)

	def __len__(self):
		# FIXME: returning zero will disable the progress bar
		return 0

	def close(self):
		if self._file is not None:
			self._file.close()
			self._file = None

	def decode(self, st: bytes) -> str:
		return st.decode(self._encoding, errors="replace")

	def getChunkSize(self, pos):
		plus = self._buf[pos:pos + 12].find(b"<d:entry")
		if plus < 1:
			return 0, 0
		bs = self._buf[pos:pos + plus]
		if plus < 4:
			bs = b"\x00" * (4 - plus) + bs
		try:
			chunkSize, = unpack("i", bs)
		except Exception as e:
			log.error(f"{self._buf[pos:pos+100]}")
			raise e
		return chunkSize, plus

	def _readEntry(self, pos: int) -> Tuple["BaseEntry", int]:
		"""
			returns (entry, pos)
		"""
		chunkSize, plus = self.getChunkSize(pos)
		pos += plus
		if chunkSize == 0:
			endI = self._buf[pos:].find(b"</d:entry>")
			if endI == -1:
				chunkSize = len(self._buf) - pos
			else:
				chunkSize = endI + 10
		entryFull = self.decode(self._buf[pos:pos + chunkSize])
		entryFull = entryFull.strip()
		if not entryFull:
			pos += chunkSize
			return None, pos
		try:
			entryRoot = etree.fromstring(entryFull)
		except etree.XMLSyntaxError as e:
			log.error(f"\n{self._buf[pos-plus:pos+100]}")
			log.error(
				f"chunkSize={chunkSize}, plus={plus}, pos={pos}, len(buf)={len(self._buf)}"
			)
			log.error(f"entryFull={entryFull!r}")
			raise e
		entryElems = entryRoot.xpath("/d:entry", namespaces=entryRoot.nsmap)
		if not entryElems:
			return None, pos
		word = entryElems[0].xpath("./@d:title", namespaces=entryRoot.nsmap)[0]
		if self._html:
			defi = self.decode(etree.tostring(entryElems[0]))
		else:
			defi = "".join([
				self.decode(etree.tostring(child))
				for child in entryElems[0].iterdescendants()
			])
		pos += chunkSize
		if self._limit <= 0:
			raise ValueError(f"self._limit = {self._limit}")
		return self._glos.newEntry(
			word, defi,
			defiFormat=self._defiFormat,
			byteProgress=(self._absPos, self._limit),
		), pos

	def __iter__(self):
		file = self._file
		limit = self._limit
		while True:
			self._absPos = file.tell()
			if self._absPos >= limit:
				break
			bufSizeB = file.read(4)  # type: bytes
			# alternative for buf, bufSize is calculated
			# ~ flag = f.tell()
			# ~ bufSize = 0
			# ~ while True:
			# ~ 	zipp = f.read(bufSize)
			# ~		try:
			# ~			# print(zipp)
			# ~			input(zipp.decode(self._encoding))
			# ~			buf = decompress(zipp[8:])
			# ~			# print(buf)
			# ~			break
			# ~		except:
			# ~			print(bufSize)
			# ~			f.seek(flag)
			# ~			bufSize = bufSize+1

			bufSize, = unpack("i", bufSizeB)  # type: int
			self._buf = decompress(file.read(bufSize)[8:])

			pos = 0
			while pos < len(self._buf):
				entry, pos = self._readEntry(pos)
				if entry is not None:
					yield entry



# -*- coding: utf-8 -*-
# appledict/indexes/__init__.py
#
# Copyright (C) 2016 Ratijas <ratijas.t@me.com>
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
"""
extended indexes generation with respect to source language.
"""

import os
import pkgutil

from pyglossary.plugins.formats_common import log

__all__ = ["languages", "log"]

languages = {}
"""
Dict[str, Callable[[Sequence[str], str], Sequence[str]]]

submodules must register languages by adding (language name -> function)
pairs to the mapping.

function must follow signature bellow:
	:param titles: flat iterable of title and altenrative titles
	:param content: cleaned entry content
	:return: iterable of indexes (str).

use
```
	from . import languages
	# or
	from appledict.indexes import languages
```
"""

here = os.path.dirname(os.path.abspath(__file__))

for _, module, _ in pkgutil.iter_modules([here]):
	try:
		__import__("%s.%s" % (__name__, module))
	except ImportError:
		log.exception(
			"error while importing indexes plugin %s" % module
		)



# -*- coding: utf-8 -*-
# xdxf/__init__.py
"""xdxf file format reader and utils to convert xdxf to html."""
#
# Copyright (C) 2016 Ratijas <ratijas.t@me.com>
#
# some parts of this file include code from:
# Aard Dictionary Tools <http://aarddict.org>.
# Copyright (C) 2008-2009  Igor Tkach
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

from os import path

from formats_common import *

enable = True
format = "Xdxf"
description = "XDXF"
extensions = [".xdxf", ".xml"]
optionsProp = {}
depends = {
	"lxml": "lxml",
}

etree = None
XML = None
tostring = None
transform = None

try:
	ModuleNotFoundError
except NameError:
	ModuleNotFoundError = ImportError

def import_xml_stuff():
	try:
		from lxml import etree as _etree
	except ModuleNotFoundError as e:
		e.msg += ", run `sudo pip3 install lxml` to install"
		raise e
	global etree, XML, tostring
	etree = _etree
	XML = etree.XML
	tostring = etree.tostring


def read(glos: GlossaryType, filename: str):
	"""
	new format
	<xdxf ...>
		<meta_info>
			All meta information about the dictionary: its title, author etc.
		</meta_info>
		<lexicon>
			<ar>article 1</ar>
			<ar>article 2</ar>
			<ar>article 3</ar>
			<ar>article 4</ar>
			...
		</lexicon>
	</xdxf>

	old format
	<xdxf ...>
		<full_name>...</full_name>
		<description>...</description>
		<ar>article 1</ar>
		<ar>article 2</ar>
		<ar>article 3</ar>
		<ar>article 4</ar>
		...
	</xdxf>
	"""
	# <!DOCTYPE xdxf SYSTEM "http://xdxf.sourceforge.net/xdxf_lousy.dtd">
	import_xml_stuff()

	with open(filename, "rb") as f:
		xdxf = XML(f.read())

	if len(xdxf) == 2:
		# new format
		read_metadata_new(glos, xdxf)
		read_xdxf_new(glos, xdxf)
	else:
		# old format
		read_metadata_old(glos, xdxf)
		read_xdxf_old(glos, xdxf)


def read_metadata_old(glos: GlossaryType, xdxf):
	full_name = xdxf.find("full_name").text
	description = xdxf.find("description").text
	if full_name:
		glos.setInfo("name", full_name)
	if description:
		glos.setInfo("description", description)


def read_xdxf_old(glos: GlossaryType, xdxf):
	add_articles(glos, xdxf.iterfind("ar"))


def read_metadata_new(glos: GlossaryType, xdxf):
	meta_info = xdxf.find("meta_info")
	if meta_info is None:
		raise ValueError("meta_info not found")

	title = meta_info.find("full_title").text
	if not title:
		title = meta_info.find("title").text
	description = meta_info.find("description").text

	if title:
		glos.setInfo("name", title)
	if description:
		glos.setInfo("description", description)


def read_xdxf_new(glos: GlossaryType, xdxf):
	add_articles(glos, xdxf.find("lexicon").iterfind("ar"))


def add_articles(glos: GlossaryType, articles):
	"""

	:param articles: iterator on <ar> tags
	:return: None
	"""
	glos.setDefaultDefiFormat("x")
	for article in articles:
		article.tail = None
		defi = tostring(article, encoding="utf-8")
		# <ar>...</ar>
		defi = defi[4:-5].strip()
		glos.addEntry(
			[toStr(w) for w in titles(article)],
			toStr(defi),
		)


def titles(article):
	"""

	:param article: <ar> tag
	:return: (title (str) | None, alternative titles (set))
	"""
	from itertools import combinations
	titles = []
	for title_element in article.findall("k"):
		n_opts = len([c for c in title_element if c.tag == "opt"])
		if n_opts:
			for j in range(n_opts + 1):
				for comb in combinations(list(range(n_opts)), j):
					titles.append(_mktitle(title_element, comb))
		else:
			titles.append(_mktitle(title_element))

	return titles


def _mktitle(title_element, include_opts=()):
	title = title_element.text
	opt_i = -1
	for c in title_element:
		if c.tag == "nu" and c.tail:
			if title:
				title += c.tail
			else:
				title = c.tail
		if c.tag == "opt":
			opt_i += 1
			if opt_i in include_opts:
				if title:
					title += c.text
				else:
					title = c.text
			if c.tail:
				if title:
					title += c.tail
				else:
					title = c.tail
	return title.strip()


def xdxf_init():
	"""
	call this only once, before `xdxf_to_html`.
	"""
	global transform

	import_xml_stuff()

	xsl = path.join(path.dirname(__file__), "xdxf.xsl")
	with open(xsl, "r") as f:
		xslt_root_txt = f.read()

	xslt_root = etree.XML(xslt_root_txt)
	transform = etree.XSLT(xslt_root)


def xdxf_to_html(xdxf_text):
	"""
	make sure to call `xdxf_init()` first.

	:param xdxf_text: xdxf formatted string
	:return: html formatted string
	"""
	from io import StringIO
	xdxf_txt = "<ar>%s</ar>" % xdxf_text
	f = StringIO(xdxf_txt)
	doc = etree.parse(f)
	result_tree = transform(doc)
	return tostring(result_tree, encoding="utf-8")



# -*- coding: utf-8 -*-

from formats_common import *

enable = True
format = "Freedict"
description = "FreeDict (tei)"
extensions = [".tei"]
optionsProp = {}
depends = {}


def write(glos: GlossaryType, filename: str):
	fp = open(filename, "w")
	title = glos.getInfo("title")

	fp.write(f"""<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE TEI.2 PUBLIC "-//TEI P3//DTD Main Document Type//EN"
"/usr/share/sgml/tei-3/tei2.dtd" [
<!ENTITY %% TEI.dictionaries "INCLUDE" > ]>
<tei.2>
<teiHeader>
<fileDesc>
<titleStmt>
	<title>{title}</title>
	<respStmt><resp>converted with</resp><name>PyGlossary</name></respStmt>
</titleStmt>
<publicationStmt><p>freedict.de</p></publicationStmt>
<sourceDesc><p>{filename}</p></sourceDesc>
</fileDesc>
</teiHeader>
<text><body>""")

	for entry in glos:
		if entry.isData():
			# FIXME
			continue
		word = entry.getWord()
		defi = entry.getDefi()
		fp.write(f"""<entry>
<form><orth>{word}</orth></form>
<gramgrp><pos>n</pos></gramgrp>
<trans><tr>{defi}</tr></trans>
</entry>""")
	fp.write("</body></text></tei.2>")
	fp.close()



# -*- coding: utf-8 -*-
# flawless_dsl/tests.py
#
# Copyright (C) 2016 Ratijas <ratijas.t@me.com>
# Copyright (C) 2016 Saeed Rasooli <saeed.gnu@gmail.com>
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
"""
test everything.
"""


import unittest

import os
import sys
from functools import partial


sys.path.insert(0, os.path.dirname(os.path.dirname(__file__)))

from flawless_dsl import layer, tag
from flawless_dsl.main import (
	process_closing_tags,
	FlawlessDSLParser,
	parse,
	BRACKET_L,
	BRACKET_R,
)

tag_i = tag.Tag('i', 'i')
tag_m = tag.Tag('m1', 'm')
tag_p = tag.Tag('p', 'p')
tag_s = tag.Tag('s', 's')


class LayerTestCase(unittest.TestCase):
	def setUp(self):
		pass

	def test_new_layer(self):
		stack = []
		l = layer.Layer(stack)
		self.assertEqual(1, len(stack))
		self.assertEqual(l, stack[0])

	def test_was_opened_AND_close_tags(self):
		stack = []
		l1, l2 = layer.Layer(stack), layer.Layer(stack)
		l1.text = '...'
		l2.tags, l2.text = {tag_i}, ',,,'

		self.assertTrue(tag.was_opened(stack, tag_i))
		self.assertFalse(tag.was_opened(stack, tag.Tag('c green', 'c')))

		layer.close_tags(stack, {tag_i}, len(stack) - 1)

		expected = []
		l = layer.Layer(expected)
		l.text = '...[i],,,[/i]'
		self.assertEqual(expected, stack)

	def test_close_layer(self):
		stack = []
		l1, l2, l3 = layer.Layer(stack), layer.Layer(stack), layer.Layer(stack)
		l1.tags, l1.text = {tag_m}, '...'
		l2.tags, l2.text = {tag_i}, ',,,'
		l3.tags, l3.text = {tag_p, tag_s}, '+++'

		expected = []
		l1, l2 = layer.Layer(expected), layer.Layer(expected)
		l1.tags, l1.text = {tag_m}, '...'
		l2.tags, l2.text = {tag_i}, ',,,[%s][%s]+++[/%s][/%s]' % (
			tag_p.opening, tag_s.opening, tag_s.closing, tag_p.closing)

		layer.close_layer(stack)
		self.assertEqual(expected, stack)


class CanonicalOrderTestCase(unittest.TestCase):
	def setUp(self):
		pass

	def test_no_tags(self):
		tags = {}
		expected = []
		result = tag.canonical_order(tags)
		self.assertEqual(expected, result)

	def test_one_tag_not_predefined(self):
		tags = {tag_p}
		expected = [tag_p]
		result = tag.canonical_order(tags)
		self.assertEqual(expected, result)

	def test_one_tag_predefined(self):
		tags = {tag_i}
		expected = [tag_i]
		result = tag.canonical_order(tags)
		self.assertEqual(expected, result)

	def test_many_tags_not_predefined(self):
		tags = {tag_p, tag_s}
		expected = [tag_p, tag_s]
		result = tag.canonical_order(tags)
		self.assertEqual(expected, result)

	def test_many_tags_predefined(self):
		tags = {tag_m, tag_p}
		expected = [tag_m, tag_p]
		result = tag.canonical_order(tags)
		self.assertEqual(expected, result)

	def test_many_tags_mixed(self):
		tags = {tag_m, tag_i, tag_s, tag_p}
		expected = [tag_m, tag_i, tag_p, tag_s]
		result = tag.canonical_order(tags)
		self.assertEqual(expected, result)


class ProcessClosingTagsTestCase(unittest.TestCase):
	def setUp(self):
		pass

	def test_index_of_layer_containing_tag(self):
		stack = []
		l1, l2, l3 = layer.Layer(stack), layer.Layer(stack), layer.Layer(stack)
		l1.tags, l1.text = {tag_m}, '...'
		l2.tags, l2.text = {tag_i, tag_s}, ',,,'
		l3.tags, l3.text = {tag_p}, '---'

		fn = partial(tag.index_of_layer_containing_tag, stack)
		self.assertEqual(0, fn(tag_m.closing))
		self.assertEqual(1, fn(tag_i.closing))
		self.assertEqual(1, fn(tag_s.closing))
		self.assertEqual(2, fn(tag_p.closing))

	def test_close_one(self):
		stack = []
		l1, l2 = layer.Layer(stack), layer.Layer(stack)
		l1.tags, l1.text = (), '...'
		l2.tags, l2.text = {tag_p}, ',,,'

		expected = []
		l = layer.Layer(expected)
		l.tags, l.text = (), '...[%s],,,[/%s]' % tag_p

		closings = {tag_p.closing}
		process_closing_tags(stack, closings)
		self.assertEqual(expected, stack)


class PutBracketsAwayTestCase(unittest.TestCase):
	def setUp(self):
		tags = frozenset({
			'b',
			'\'',
			'c',
			'i',
			'sup',
			'sub',
			'ex',
			'p',
			'*',
			('m', '\d'),
		})
		parser = FlawlessDSLParser(tags)
		self.put_brackets_away = parser.put_brackets_away

	def testStandaloneLeftEscapedAtTheBeginning(self):
		before = "[..."
		after = "%s..." % BRACKET_L
		self.assertEqual(after, self.put_brackets_away(before))

	def testStandaloneRightEscapedAtTheBeginning(self):
		before = "]..."
		after = "%s..." % BRACKET_R
		self.assertEqual(after, self.put_brackets_away(before))

	def testStandaloneLeftEscaped(self):
		before = r"...\[,,,"
		after = r"...\%s,,," % BRACKET_L
		self.assertEqual(after, self.put_brackets_away(before))

	def testStandaloneRightEscaped(self):
		before = r"...\],,,"
		after = r"...\%s,,," % BRACKET_R
		self.assertEqual(after, self.put_brackets_away(before))

	def testStandaloneLeftNonEscaped(self):
		before = "...[,,,"
		after = "...%s,,," % BRACKET_L
		self.assertEqual(after, self.put_brackets_away(before))

	def testStandaloneRightNonEscaped(self):
		before = "...],,,"
		after = "...%s,,," % BRACKET_R
		self.assertEqual(after, self.put_brackets_away(before))

	def testStandaloneLeftNonEscapedBeforeTagName(self):
		before = "...[p ,,,"
		after = "...%sp ,,," % BRACKET_L
		self.assertEqual(after, self.put_brackets_away(before))

	def testStandaloneRightNonEscapedAfterTagName(self):
		before = "c]..."
		after = "c%s..." % BRACKET_R
		self.assertEqual(after, self.put_brackets_away(before))

	def testPairEscaped(self):
		before = r"...\[the\],,,"
		after = r"...\%sthe\%s,,," % (BRACKET_L, BRACKET_R)
		self.assertEqual(after, self.put_brackets_away(before))

	def testPairEscapedAroundTagName(self):
		before = r"...\[i\],,,"
		after = r"...\%si\%s,,," % (BRACKET_L, BRACKET_R)
		self.assertEqual(after, self.put_brackets_away(before))

	def testPairEscapedAroundClosingTagName(self):
		before = r"...\[/i\],,,"
		after = r"...\%s/i\%s,,," % (BRACKET_L, BRACKET_R)
		self.assertEqual(after, self.put_brackets_away(before))

	def testMixed(self):
		before = r"[i]...\[on \]\[the] to[p][/i]"
		after = r"[i]...\{L}on \{R}\{L}the{R} to[p][/i]".format(
			L=BRACKET_L,
			R=BRACKET_R,
		)
		self.assertEqual(after, self.put_brackets_away(before))

	def testEverythingEscaped(self):
		before = """ change it to \[b\]...\[c\]...\[/c\]\[/b\]\[c\]...\[/c\]"""
		after = before
		self.assertEqual(after, parse(before))


class FlawlessDSLParserTestCase(unittest.TestCase):
	def setUp(self):
		self.split_join = lambda x: FlawlessDSLParser.join_paragraphs(
			*FlawlessDSLParser.split_line_by_paragraphs(x))

	def testStartsWithStandaloneClosed(self):
		before = """[/p]..."""
		after = """..."""
		self.assertEqual(after, parse(before))

	def testStandaloneClosedAtTheBeginning(self):
		before = """...[/p],,,"""
		after = """...,,,"""
		self.assertEqual(after, parse(before))

	def testStandaloneClosedAtTheBeginningBeforeMarkup(self):
		before = """...[/p],,,[i][b]+++[/b][/i]---"""
		after = """...,,,[i][b]+++[/b][/i]---"""
		self.assertEqual(after, parse(before))

	def testEndsWithStandaloneOpened(self):
		before = """...[i]"""
		after = """..."""
		self.assertEqual(after, parse(before))

	def testStandaloneOpenedAtTheEnd(self):
		before = """...[i],,,"""
		after = """...,,,"""
		self.assertEqual(after, parse(before))

	def testStandaloneOpenedAtTheEndAfterMarkup(self):
		before = """...[i][b],,,[/b][/i]+++[i]---"""
		after = """...[i][b],,,[/b][/i]+++---"""
		self.assertEqual(after, parse(before))

	def testWrongOrder2(self):
		before = """...[i][b],,,[/i][/b]+++"""
		after = """...[i][b],,,[/b][/i]+++"""
		self.assertEqual(after, parse(before))

	def testWrongOrder3(self):
		before = """...[i][c],,,[b]+++[/i][/c][/b]---"""
		after = """...[p],,,[b]+++[/b][/p]---"""
		self.assertEqual(after, parse(before))

	def testOpenOneCloseAnother(self):
		before = """...[i],,,[/p]+++"""
		after = """...,,,+++"""
		self.assertEqual(after, parse(before))

	def testStartsWtihClosingAndEndsWithOpening(self):
		before = """[/c]...[i]"""
		after = """..."""
		self.assertEqual(after, parse(before))

	def testValidEmptyTagsDestructionOne(self):
		before = """...[i][/i],,,"""
		after = """...,,,"""
		self.assertEqual(after, parse(before))

	def testValidEmptyTagsDestructionMany(self):
		before = """...[b][c][i][/i][/c][/b],,,"""
		after = """...,,,"""
		self.assertEqual(after, parse(before))

	def testBrokenEmptyTagsDestructionMany(self):
		before = """...[b][i][c][/b][/c][/i],,,"""
		after = """...,,,"""
		self.assertEqual(after, parse(before))

	def testNestedWithBrokenOutter(self):
		before = """[i][p]...[/p][/c]"""
		after = """[p]...[/p]"""
		self.assertEqual(after, parse(before))

	def testHorriblyBrokenTags(self):
		before = """[/c]...[i][/p],,,[/i]+++[b]"""
		after = """...[i],,,[/i]+++"""
		self.assertEqual(after, parse(before))

	def testWrongOrder2_WithConent(self):
		before = """[b]...[c red]...[/b]...[/c]"""
		after = """[b]...[c red]...[/c][/b][c red]...[/c]"""
		self.assertEqual(after, parse(before))

	def testWrongOrderWithTextBefore(self):
		before = "[c]...[i],,,[/c][/i]"
		after = "[c]...[i],,,[/i][/c]"
		self.assertEqual(after, parse(before))

	def testRespect_m_TagsProperly(self):
		before = """ [m1]for tags like: [p]n[/c][/i][/p], the line needs scan again[/m]"""
		after = """ [m1]for tags like: [p]n[/p], the line needs scan again[/m]"""
		self.assertEqual(after, parse(before))

	def testNoTagsDoNothing(self):
		before = after = """no tags, do nothing"""
		self.assertEqual(after, parse(before))

	def testValidNestedTags(self):
		before = """...[i][c][b]...[/b][/c][/i]..."""
		after = """...[b][p]...[/p][/b]..."""
		self.assertEqual(after, parse(before))

	def testBrokenNestedTags(self):
		before = """...[b][i][c]...[/b][/c][/i]..."""
		after = """...[b][p]...[/p][/b]..."""
		self.assertEqual(after, parse(before))

	def testEscapedBrackets(self):
		before = after = r"""on \[the\] top"""
		self.assertEqual(after, parse(before))

	def testPoorlyEscapedBracketsWithTags(self):
		before = r"""...\[c],,,[/c]+++"""
		after = r"""...\[c],,,+++"""
		self.assertEqual(after, parse(before))

	def testPoorlyEscapedBracketsWithTags2(self):
		before = r"""on \[the\] [b]roof[/b]]"""
		after = r"""on \[the\] [b]roof[/b]]"""
		self.assertEqual(after, parse(before))

	def testValidRealDictionaryArticle(self):
		# zh => ru, http://bkrs.info/slovo.php?ch=和田
		before = after = """和田
[m1][p]г. и уезд[/p] Хотан ([i]Синьцзян-Уйгурский[c] авт.[/c] р-н, КНР[/i])[/m]\
[m2][*][ex]和田玉 Хотанский нефрит[/ex][/*][/m]"""
		self.assertEqual(after, parse(before))

	def testBrokenRealDictionaryArticle(self):
		# zh => ru, http://bkrs.info/slovo.php?ch=一一相应
		before = """一一相应
yīyī xiāngyìng
[m1][c][i]мат.[/c][/i] взаимнооднозначное соответствие[/m]"""
		after = """一一相应
yīyī xiāngyìng
[m1][p]мат.[/p] взаимнооднозначное соответствие[/m]"""
		self.assertEqual(after, parse(before))

	def testBrokenManyRealDictionaryArticle(self):
		# zh => ru, http://bkrs.info/slovo.php?ch=一轮
		before = """一轮
yīlún
[m1]1) одна очередь[/m][m1]2) цикл ([i]в 12 лет[/i])[/m][m1]3) диск ([c][i]напр.[/c] луны[/i])[/m]\
[m1]4) [c] [i]спорт[/c][/i] раунд, круг ([i]встречи спортсменов[/i])[/m]\
[m1]5) [c] [i]дипл.[/c][/i] раунд ([i]переговоров[/i])[/m]"""
		after = """一轮
yīlún
[m1]1) одна очередь[/m][m1]2) цикл ([i]в 12 лет[/i])[/m][m1]3) диск ([i][c]напр.[/c] луны[/i])[/m]\
[m1]4) [c] [i]спорт[/i][/c] раунд, круг ([i]встречи спортсменов[/i])[/m]\
[m1]5) [c] [i]дипл.[/i][/c] раунд ([i]переговоров[/i])[/m]"""
		self.assertEqual(after, parse(before))

	def testSameTagsNested(self):
		before = "...[p],,,[p]+++[/p]---[/p]```"
		after = "...[p],,,+++[/p]---```"
		self.assertEqual(after, parse(before))

	def testOneLastTextLetter(self):
		before = after = "b"
		self.assertEqual(after, parse(before))

	def testOneLastTextLetterAfterTag(self):
		before = after = "...[b],,,[/b]b"
		self.assertEqual(after, parse(before))

	def testTagMInsideAnotherTag(self):
		# tag order.
		before = "[c][m1]...[/m][/c]"
		after = "[m1][c]...[/c][/m]"
		self.assertEqual(after, parse(before))

	def testTagMInsideAnotherTagAfterText(self):
		before = "[c]...[m1],,,[/m][/c]"
		after = "[c]...[/c][m1][c],,,[/c][/m]"
		self.assertEqual(after, parse(before))

	def testTagMDeepInside(self):
		before = "...[i],,,[b]+++[c green][/b]---[m1]```[/i][/c][/m]..."
		after = "...[i],,,[b]+++[/b][c green]---[/c][/i][m1][i][c green]```[/c][/i][/m]..."
		self.assertEqual(after, parse(before))

	def testTagMInsideBroken(self):
		before = "[m1][*]- [ref]...[/ref][/m][m1]- [ref],,,[/ref][/*][/m]"
		after = "[m1][*]- [ref]...[/ref][/*][/m][m1][*]- [ref],,,[/ref][/*][/m]"
		self.assertEqual(after, parse(before))


if __name__ == '__main__':
	unittest.main()



import unittest
import random
from typing import Optional, Tuple, List, Any

from pyglossary.option import *


class TestOptionValidateBoolNumber(unittest.TestCase):
	def caseOK(self, cls, raw: str, value: Optional[bool]):
		opt = cls()
		valueActual, ok = opt.evaluate(raw)
		self.assertTrue(ok, "evaluate failed")
		self.assertEqual(valueActual, value)
		ok2 = opt.validate(valueActual)
		self.assertEqual(ok2, True, "validate failed")

	def caseFailed(self, cls, raw: str, value: Optional[bool]):
		opt = cls()
		valueActual, ok = opt.evaluate(raw)
		self.assertFalse(ok)
		self.assertEqual(valueActual, value)

	def test_bool_ok(self):
		self.caseOK(BoolOption, "True", True)
		self.caseOK(BoolOption, "False", False)

		self.caseOK(BoolOption, "true", True)
		self.caseOK(BoolOption, "false", False)

		self.caseOK(BoolOption, "TRUE", True)
		self.caseOK(BoolOption, "FALSE", False)

		self.caseOK(BoolOption, "1", True)
		self.caseOK(BoolOption, "0", False)

		self.caseOK(BoolOption, "yes", True)
		self.caseOK(BoolOption, "no", False)

		self.caseOK(BoolOption, "YES", True)
		self.caseOK(BoolOption, "NO", False)

	def test_bool_failed(self):
		self.caseFailed(BoolOption, "Y", None)
		self.caseFailed(BoolOption, "N", None)
		self.caseFailed(BoolOption, "YESS", None)
		self.caseFailed(BoolOption, "NOO", None)
		self.caseFailed(BoolOption, "123", None)
		self.caseFailed(BoolOption, "a", None)

	def test_int_ok(self):
		self.caseOK(IntOption, "0", 0)
		self.caseOK(IntOption, "1", 1)
		self.caseOK(IntOption, "-1", -1)
		self.caseOK(IntOption, "1234", 1234)

	def test_int_failed(self):
		self.caseFailed(IntOption, "abc", None)
		self.caseFailed(IntOption, "12f", None)
		self.caseFailed(IntOption, "fff", None)

	def test_float_ok(self):
		self.caseOK(FloatOption, "0", 0.0)
		self.caseOK(FloatOption, "1", 1.0)
		self.caseOK(FloatOption, "-1", -1.0)
		self.caseOK(FloatOption, "1234", 1234.0)
		self.caseOK(FloatOption, "1.5", 1.5)
		self.caseOK(FloatOption, "-7.9", -7.9)

	def test_float_failed(self):
		self.caseFailed(FloatOption, "abc", None)
		self.caseFailed(FloatOption, "12f", None)
		self.caseFailed(FloatOption, "fff", None)


class TestOptionValidateStr(unittest.TestCase):
	def newTester(self, customValue: bool, values: List[str]):
		def test(raw: str, valid: bool):
			opt = StrOption(customValue=customValue, values=values)
			valueActual, evalOkActual = opt.evaluate(raw)
			self.assertEqual(evalOkActual, True, "evaluate failed")
			self.assertEqual(valueActual, raw)
			validActual = opt.validate(valueActual)
			self.assertEqual(validActual, valid, "validate failed")
		return test

	def test_1(self):
		test = self.newTester(False, ["a", "b", "c"])
		test("a", True)
		test("b", True)
		test("c", True)
		test("d", False)
		test("123", False)

	def test_2(self):
		test = self.newTester(True, ["a", "b", "3"])
		test("a", True)
		test("b", True)
		test("c", True)
		test("d", True)
		test("123", True)


class TestOptionValidateDict(unittest.TestCase):
	def caseOK(self, raw: str, value: Optional[Dict]):
		opt = DictOption()
		valueActual, ok = opt.evaluate(raw)
		self.assertTrue(ok, "evaluate failed")
		self.assertEqual(valueActual, value)
		ok2 = opt.validate(valueActual)
		self.assertEqual(ok2, True, "validate failed")

	def caseEvalFail(self, raw: str):
		opt = DictOption()
		valueActual, ok = opt.evaluate(raw)
		self.assertFalse(ok)
		self.assertEqual(valueActual, None)

	def test_dict_ok(self):
		self.caseOK("", None)
		self.caseOK("{}", {})
		self.caseOK('{"a": 1}', {"a": 1})
		self.caseOK('{"a": "b", "123":456}', {"a": "b", "123": 456})

	def test_dict_syntaxErr(self):
		self.caseEvalFail("123abc")
		self.caseEvalFail('{')
		self.caseEvalFail("(")
		self.caseEvalFail('{"a": 1')
		self.caseEvalFail('{"a": 1]')
		self.caseEvalFail('][')

	def test_dict_notDict(self):
		self.caseEvalFail("123")
		self.caseEvalFail("[]")
		self.caseEvalFail("[1, 2, 3]")
		self.caseEvalFail('["a", 2, 3.5]')
		self.caseEvalFail('{10, 20, 30}')


class TestOptionValidateList(unittest.TestCase):
	def caseOK(self, raw: str, value: Optional[Dict]):
		opt = ListOption()
		valueActual, ok = opt.evaluate(raw)
		self.assertTrue(ok, "evaluate failed")
		self.assertEqual(valueActual, value)
		ok2 = opt.validate(valueActual)
		self.assertEqual(ok2, True, "validate failed")

	def caseEvalFail(self, raw: str):
		opt = ListOption()
		valueActual, ok = opt.evaluate(raw)
		self.assertFalse(ok, "evaluale did not fail, valueActual=%r" % valueActual)
		self.assertEqual(valueActual, None)

	def test_list_ok(self):
		self.caseOK("", None)
		self.caseOK("[]", [])
		self.caseOK('["a", "b"]', ["a", "b"])
		self.caseOK("[1, 2, 3]", [1, 2, 3])
		self.caseOK('["a", 2, 3.5]', ["a", 2, 3.5])

	def test_list_syntaxErr(self):
		self.caseEvalFail("123abc")
		self.caseEvalFail('{')
		self.caseEvalFail("(")
		self.caseEvalFail('{"a": 1')
		self.caseEvalFail('{"a": 1]')
		self.caseEvalFail('][')

	def test_list_notList(self):
		self.caseEvalFail("123")
		self.caseEvalFail('{10, 20, 30}')
		self.caseEvalFail('{"a": 1}')
		self.caseEvalFail('{"a": "b", "123":456}')


if __name__ == "__main__":
	unittest.main()



# -*- coding: utf-8 -*-
# appledict/__init__.py
# Output to Apple Dictionary xml sources for Dictionary Development Kit.
#
# Copyright (C) 2016 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# Copyright (C) 2016 Ratijas <ratijas.t@me.com>
# Copyright (C) 2012-2015 Xiaoqiang Wang <xiaoqiangwang AT gmail DOT com>
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

import sys
import os
from os.path import abspath, basename

import re
import pkgutil
import shutil

from typing import TextIO

from pyglossary.plugins.formats_common import *
from ._dict import *

import xdxf

sys.setrecursionlimit(10000)

enable = True
format = "AppleDict"
description = "AppleDict Source (xml)"
extensions = [".xml"]
# FIXME: rename indexes arg/option to indexes_lang?
optionsProp = {
	"cleanHTML": BoolOption(comment="use BeautifulSoup parser"),
	"css": StrOption(comment="custom .css file path"),
	"xsl": StrOption(comment="custom XSL transformations file path"),
	"defaultPrefs": DictOption(
		comment='default prefs in python dict format',
		# example: {"key": "value", "version": "1"}
	),
	"prefsHTML": StrOption(comment="preferences XHTML file path"),
	"frontBackMatter": StrOption(comment="XML file path with top-level tag"),
	"jing": BoolOption(comment="run Jing check on generated XML"),
	"indexes": StrOption(customValue=False, values=["", "ru", "zh"]),
}
depends = {
	"lxml": "lxml",
	"bs4": "beautifulsoup4",
	"html5lib": "html5lib",
}

BeautifulSoup = None

def loadBeautifulSoup():
	global BeautifulSoup
	try:
		import bs4 as BeautifulSoup
	except:
		try:
			import BeautifulSoup
		except:
			return
	if int(BeautifulSoup.__version__.split(".")[0]) < 4:
		raise ImportError(
			"BeautifulSoup is too old, required at least version 4, " +
			"%r found.\n" % BeautifulSoup.__version__ +
			"Please run `sudo pip3 install lxml beautifulsoup4 html5lib`"
		)

def abspath_or_None(path):
	return os.path.abspath(os.path.expanduser(path)) if path else None


def write_header(glos: GlossaryType, toFile: TextIO, frontBackMatter: Optional[str]):
	# write header
	toFile.write(
		'<?xml version="1.0" encoding="UTF-8"?>\n'
		'<d:dictionary xmlns="http://www.w3.org/1999/xhtml" '
		'xmlns:d="http://www.apple.com/DTDs/DictionaryService-1.0.rng">\n'
	)

	if frontBackMatter:
		with open(frontBackMatter, "r") as front_back_matter:
			toFile.write(front_back_matter.read())


def format_default_prefs(defaultPrefs):
	"""
	:type defaultPrefs: dict or None

	as by 14th of Jan 2016, it is highly recommended that prefs should contain
	{"version": "1"}, otherwise Dictionary.app does not keep user changes
	between restarts.
	"""
	if not defaultPrefs:
		return ""
	if not isinstance(defaultPrefs, dict):
		raise TypeError("defaultPrefs not a dictionary: %r" % defaultPrefs)
	if str(defaultPrefs.get("version", None)) != "1":
		log.error("default prefs does not contain {'version': '1'}.  prefs "
				  "will not be persistent between Dictionary.app restarts.")
	return "\n".join("\t\t<key>%s</key>\n\t\t<string>%s</string>" % i
					 for i in sorted(defaultPrefs.items())).strip()


def write_css(fname, css_file):
	with open(fname, "wb") as toFile:
		if css_file:
			with open(css_file, "rb") as fromFile:
				toFile.write(fromFile.read())
		else:
			toFile.write(pkgutil.get_data(
				__name__,
				"templates/Dictionary.css",
			))


def write(
	glos: GlossaryType,
	dirname: str,
	cleanHTML: bool = True,
	css: str = "",
	xsl: str = "",
	defaultPrefs: Optional[Dict] = None,
	prefsHTML: str = "",
	frontBackMatter: str = "",
	jing: bool = False,
	indexes: str = "",# FIXME: rename to indexes_lang?
):
	"""
	write glossary to Apple dictionary .xml and supporting files.

	:type glos: pyglossary.glossary.Glossary
	:type dirname: str, directory path, must not have extension

	:type cleanHTML: bool
	:param cleanHTML: pass True to use BeautifulSoup parser.

	:type css: str
	:param css: path to custom .css file

	:type xsl: str
	:param xsl: path to custom XSL transformations file.

	:type defaultPrefs: dict or None
	:param defaultPrefs: Default prefs in python dictionary literal format,
	i.e. {"key1": "value1", "key2": "value2", ...}.  All keys and values must
	be quoted strings; not allowed characters (e.g. single/double quotes,
	equal sign "=", semicolon) must be escaped as hex code according to
	python string literal rules.

	:type prefsHTML: str
	:param prefsHTML: path to XHTML file with user interface for dictionary's
	preferences.  refer to Apple's documentation for details.

	:type frontBackMatter: str
	:param frontBackMatter: path to XML file with top-level tag
	<d:entry id="front_back_matter" d:title="Your Front/Back Matter Title">
		your front/back matter entry content
	</d:entry>

	:type jing: bool
	:param jing: pass True to run Jing check on generated XML.

	# FIXME: rename to indexes_lang?
	:type indexes: str
	:param indexes: Dictionary.app is dummy and by default it don't know
	how to perform flexible search.  we can help it by manually providing
	additional indexes to dictionary entries.
	"""
	global BeautifulSoup

	if not isdir(dirname):
		os.mkdir(dirname)

	xdxf.xdxf_init()

	if cleanHTML:
		if BeautifulSoup is None:
			loadBeautifulSoup()
		if BeautifulSoup is None:
			log.warning(
				"cleanHTML option passed but BeautifulSoup not found.  " +
				"to fix this run `sudo pip3 install lxml beautifulsoup4 html5lib`"
			)
	else:
		BeautifulSoup = None

	fileNameBase = basename(dirname).replace(".", "_")
	filePathBase = join(dirname, fileNameBase)
	# before chdir (outside indir block)
	css = abspath_or_None(css)
	xsl = abspath_or_None(xsl)
	prefsHTML = abspath_or_None(prefsHTML)
	frontBackMatter = abspath_or_None(frontBackMatter)

	generate_id = id_generator()
	generate_indexes = indexes_generator(indexes)

	glos.setDefaultDefiFormat("h")

	myResDir = join(dirname, "OtherResources")
	if not isdir(myResDir):
		os.mkdir(myResDir)

	with open(filePathBase + ".xml", "w", encoding="utf-8") as toFile:
		write_header(glos, toFile, frontBackMatter)
		for entryI, entry in enumerate(glos):
			if entry.isData():
				entry.save(myResDir)
				continue

			words = entry.getWords()
			word, alts = words[0], words[1:]
			defi = entry.getDefi()

			long_title = _normalize.title_long(
				_normalize.title(word, BeautifulSoup)
			)
			if not long_title:
				continue

			_id = next(generate_id)
			if BeautifulSoup:
				title_attr = BeautifulSoup.dammit.EntitySubstitution\
					.substitute_xml(long_title, True)
			else:
				title_attr = str(long_title)

			content_title = long_title
			if entry.getDefiFormat() == "x":
				defi = xdxf.xdxf_to_html(defi)
				content_title = None
			content = format_clean_content(content_title, defi, BeautifulSoup)

			toFile.write(
				'<d:entry id="%s" d:title=%s>\n' % (_id, title_attr) +
				generate_indexes(long_title, alts, content, BeautifulSoup) +
				content +
				"\n</d:entry>\n"
			)

		toFile.write("</d:dictionary>\n")

	if xsl:
		shutil.copy(xsl, myResDir)

	if prefsHTML:
		shutil.copy(prefsHTML, myResDir)

	write_css(filePathBase + ".css", css)

	with open(join(dirname, "Makefile"), "w") as toFile:
		toFile.write(
			toStr(pkgutil.get_data(
				__name__,
				"templates/Makefile",
			)) % {"dict_name": fileNameBase}
		)

	copyright = glos.getInfo("copyright")
	if BeautifulSoup:
		# strip html tags
		copyright = str(BeautifulSoup.BeautifulSoup(
			copyright,
			features="lxml"
		).text)

	# if DCSDictionaryXSL provided but DCSDictionaryDefaultPrefs <dict/> not
	# present in Info.plist, Dictionary.app will crash.
	with open(filePathBase + ".plist", "w", encoding="utf-8") as toFile:
		toFile.write(
			toStr(pkgutil.get_data(
				__name__,
				"templates/Info.plist",
			)) % {
				"CFBundleIdentifier":
					fileNameBase.replace(" ", ""),  # identifier must be unique
				"CFBundleDisplayName":
					glos.getInfo("name"),
				"CFBundleName":
					fileNameBase,
				"DCSDictionaryCopyright":
					copyright,
				"DCSDictionaryManufacturerName":
					glos.getInfo("author"),
				"DCSDictionaryXSL":
					basename(xsl) if xsl else "",
				"DCSDictionaryDefaultPrefs":
					format_default_prefs(defaultPrefs),
				"DCSDictionaryPrefsHTML":
					basename(prefsHTML) if prefsHTML else "",
				"DCSDictionaryFrontMatterReferenceID":
					"<key>DCSDictionaryFrontMatterReferenceID</key>\n"
					"\t<string>front_back_matter</string>" if frontBackMatter
					else "",
			}
		)

	if jing:
		from .jing import run as jing_run
		jing_run(filePathBase + ".xml")



# -*- coding: utf-8 -*-
# sdict.py
# Loader engine for AXMASoft's open dictionary format
#
# Copyright (C) 2010-2016 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# Copyright (C) 2006-2008 Igor Tkach, as part of SDict Viewer:
#			   http://sdictviewer.sf.net
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

from struct import unpack
from formats_common import *

enable = True
format = "Sdict"
description = "Sdictionary Binary(dct)"
extensions = [".dct"]
optionsProp = {
	"encoding": EncodingOption(),
}
depends = {}


class GzipCompression(object):
	def __str__(self):
		return "gzip"

	def decompress(self, string):
		import zlib
		return zlib.decompress(string)


class Bzip2Compression(object):
	def __str__(self):
		return "bzip2"

	def decompress(self, string):
		import bz2
		return bz2.decompress(string)


class NoCompression(object):
	def __str__(self):
		return "no compression"

	def decompress(self, string):
		return string


compressions = [
	NoCompression(),
	GzipCompression(),
	Bzip2Compression(),
]


def read_raw(s, fe):
	return s[fe.offset:fe.offset + fe.length]


def read_str(s, fe):
	read_raw(s, fe).replace(b"\x00", b"")


def read_int(s, fe=None):
	return unpack("<I", read_raw(s, fe) if fe else s)[0]


def read_short(raw):
	return unpack("<H", raw)[0]


def read_byte(raw):
	return unpack("<B", raw)[0]


class FormatElement(object):
	def __init__(self, offset, length):
		self.offset = offset
		self.length = length


class Header(object):
	f_signature = FormatElement(0x0, 4)
	f_input_lang = FormatElement(0x4, 3)
	f_output_lang = FormatElement(0x7, 3)
	f_compression = FormatElement(0xa, 1)
	f_num_of_words = FormatElement(0xb, 4)
	f_length_of_short_index = FormatElement(0xf, 4)
	f_title = FormatElement(0x13, 4)
	f_copyright = FormatElement(0x17, 4)
	f_version = FormatElement(0x1b, 4)
	f_short_index = FormatElement(0x1f, 4)
	f_full_index = FormatElement(0x23, 4)
	f_articles = FormatElement(0x27, 4)

	def parse(self, st):
		self.signature = read_str(st, self.f_signature)
		if self.signature != b"sdct":
			raise ValueError("Not a valid sdict dictionary")
		self.word_lang = read_str(st, self.f_input_lang)
		self.article_lang = read_str(st, self.f_output_lang)
		self.short_index_length = read_int(st, self.f_length_of_short_index)
		comp_and_index_levels_byte = read_byte(
			read_raw(st, self.f_compression)
		)
		self.compressionType = comp_and_index_levels_byte & 0b1111
		self.short_index_depth = comp_and_index_levels_byte >> 4
		self.num_of_words = read_int(st, self.f_num_of_words)
		self.title_offset = read_int(st, self.f_title)
		self.copyright_offset = read_int(st, self.f_copyright)
		self.version_offset = read_int(st, self.f_version)
		self.articles_offset = read_int(st, self.f_articles)
		self.short_index_offset = read_int(st, self.f_short_index)
		self.full_index_offset = read_int(st, self.f_full_index)


class Reader(object):
	def __init__(self, glos):
		self._glos = glos
		self.clear()

	def clear(self):
		self._file = None
		self._filename = ""
		self._encoding = ""
		self._header = Header()

	def open(self, filename, encoding="utf-8"):
		self._file = open(filename, "rb")
		self._header.parse(self._file.read(43))
		self._compression = compressions[self._header.compressionType]
		self.short_index = self.readShortIndex()
		self._glos.setInfo(
			"name",
			self.readUnit(self._header.title_offset),
		)
		self._glos.setInfo(
			"version",
			self.readUnit(self._header.version_offset)
		)
		self._glos.setInfo(
			"copyright",
			self.readUnit(self._header.copyright_offset),
		)
		log.debug("SDict word count: %s", len(self))  # correct? FIXME

	def close(self):
		self._file.close()
		self.clear()

	def __len__(self):
		return self._header.num_of_words

	def readUnit(self, pos):
		f = self._file
		f.seek(pos)
		record_length = read_int(f.read(4))
		return self._compression.decompress(f.read(record_length))

	def readShortIndex(self):
		self._file.seek(self._header.short_index_offset)
		s_index_depth = self._header.short_index_depth
		index_entry_len = (s_index_depth+1)*4
		short_index_str = self._file.read(
			index_entry_len * self._header.short_index_length
		)
		short_index_str = self._compression.decompress(short_index_str)
		index_length = self._header.short_index_length
		short_index = [{} for i in range(s_index_depth+2)]
		depth_range = range(s_index_depth)
		for i in range(index_length):
			entry_start = start_index = i*index_entry_len
			short_word = ""
			try:
				for j in depth_range:
					# inlined unpack yields ~20% performance gain
					# compared to calling read_int()
					uchar_code = unpack(
						"<I",
						short_index_str[start_index:start_index+4]
					)[0]
					start_index += 4
					if uchar_code == 0:
						break
					short_word += chr(uchar_code)
			except ValueError as ve:
				# If Python is built without wide unicode support (which is
				# the case on Maemo) it may not be possible to use some
				# unicode chars. It seems best to ignore such index items
				# The rest of the dictionary should be usable.
				log.error(
					"Failed to decode short index item %s" % i +
					", will ignore: %s" % ve
				)
				continue
			pointer_start = entry_start+s_index_depth*4
			pointer = unpack(
				"<I",
				short_index_str[pointer_start:pointer_start+4]
			)[0]
			short_index[len(short_word)][short_word] = pointer
		return short_index

	def __iter__(self):
		pos = self._header.full_index_offset
		next_ptr = 0
		while True:
			pos += next_ptr
			item = self.readFullIndexItem(pos)
			if item is None:
				break
			(next_ptr, word, ptr) = item
			if word is None:
				break
			word = toStr(word)
			defi = self.readUnit(self._header.articles_offset + ptr)
			defi = toStr(defi)
			defi = defi.replace("<BR>", "\n").replace("<br>", "\n")
			yield self._glos.newEntry(word, defi)

	def readFullIndexItem(self, pointer):
		try:
			f = self._file
			f.seek(pointer)
			s = f.read(8)
			next_word = unpack("<H", s[:2])[0]
			article_pointer = unpack("<I", s[4:])[0]
			word = f.read(next_word - 8) if next_word else None
			return next_word, word, article_pointer
		except Exception as e:
			if pointer >= self._header.articles_offset:
				log.error(
					"Warning: attempt to read word from "
					"illegal position in dict file"
				)
				return None
			log.exception("")

	def readArticle(self, pointer):
		return self.readUnit(self._header.articles_offset + pointer)



# -*- coding: utf-8 -*-

from formats_common import *

enable = True
format = "GettextPo"
description = "Gettext Source (po)"
extensions = [".po"]
optionsProp = {
	"resources": BoolOption(),
}
depends = {}


class Reader(object):
	def __init__(self, glos: GlossaryType):
		self._glos = glos
		self.clear()

	def clear(self):
		self._filename = ""
		self._file = None
		self._wordCount = None
		self._resDir = ""
		self._resFileNames = []

	def open(self, filename):
		self._filename = filename
		self._file = open(filename)
		self._resDir = filename + "_res"
		if isdir(self._resDir):
			self._resFileNames = os.listdir(self._resDir)
		else:
			self._resDir = ""
			self._resFileNames = []

	def close(self):
		if self._file:
			self._file.close()
		self.clear()

	def __len__(self):
		from pyglossary.file_utils import fileCountLines
		if self._wordCount is None:
			log.debug("Try not to use len(reader) as it takes extra time")
			self._wordCount = fileCountLines(
				self._filename,
				newline="\nmsgid",
			)
		return self._wordCount

	def __iter__(self):
		from polib import unescape as po_unescape
		word = ""
		defi = ""
		msgstr = False
		wordCount = 0
		for line in self._file:
			line = line.strip()
			if not line:
				continue
			if line.startswith("#"):
				continue
			if line.startswith("msgid "):
				if word:
					yield self._glos.newEntry(word, defi)
					wordCount += 1
					word = ""
					defi = ""
				word = po_unescape(line[6:])
				msgstr = False
			elif line.startswith("msgstr "):
				if msgstr:
					log.error("msgid omitted!")
				defi = po_unescape(line[7:])
				msgstr = True
			else:
				if msgstr:
					defi += po_unescape(line)
				else:
					word += po_unescape(line)
		if word:
			yield self._glos.newEntry(word, defi)
			wordCount += 1
		self._wordCount = wordCount


def write(glos: GlossaryType, filename: str, resources: bool = True):
	from polib import escape as po_escape
	with open(filename, "w") as toFile:
		toFile.write("#\nmsgid ""\nmsgstr ""\n")
		for key, value in glos.iterInfo():
			toFile.write('"%s: %s\\n"\n' % (key, value))
		for entry in glos:
			if entry.isData():
				if resources:
					entry.save(filename + "_res")
				continue
			word = entry.getWord()
			defi = entry.getDefi()
			toFile.write("msgid %s\nmsgstr %s\n\n" % (
				po_escape(word),
				po_escape(defi),
			))



# -*- coding: utf-8 -*-

from typing import (
	List,
	Tuple,
	Callable,
	Union,
	Any,
	Optional,
)

MultiStr = Union[str, List[str]]

RawEntryType = Union[
	Tuple[MultiStr, MultiStr],
	Tuple[MultiStr, MultiStr, str],
]


class BaseEntry(object):
	def isData(self) -> bool:
		raise NotImplementedError

	def getFileName(self) -> str:
		raise NotImplementedError

	def getData(self) -> bytes:
		raise NotImplementedError

	def save(self, directory: str) -> str:
		raise NotImplementedError

	def getWord(self) -> str:
		raise NotImplementedError

	def getWords(self) -> List[str]:
		raise NotImplementedError

	def getDefi(self) -> str:
		raise NotImplementedError

	def getDefis(self) -> List[str]:
		raise NotImplementedError

	def getDefiFormat(self) -> str:
		# TODO: type: Literal["m", "h", "x", "b"]
		raise NotImplementedError

	def setDefiFormat(self, defiFormat: str) -> None:
		# TODO: type: Literal["m", "h", "x", "b"]
		raise NotImplementedError

	def detectDefiFormat(self) -> None:
		raise NotImplementedError

	def addAlt(self, alt: str) -> None:
		raise NotImplementedError

	def editFuncWord(self, func: Callable[[str], str]) -> None:
		raise NotImplementedError

	def editFuncDefi(self, func: Callable[[str], str]) -> None:
		raise NotImplementedError

	def strip(self) -> None:
		raise NotImplementedError

	def replaceInWord(self, source: str, target: str) -> None:
		raise NotImplementedError

	def replaceInDefi(self, source: str, target: str) -> None:
		raise NotImplementedError

	def replace(self, source: str, target: str) -> None:
		raise NotImplementedError

	def getRaw(self) -> RawEntryType:
		raise NotImplementedError

	@staticmethod
	def getEntrySortKey(
		key: Optional[Callable[[str], Any]] = None,
	) -> Callable[["BaseEntry"], Any]:
		raise NotImplementedError

	@staticmethod
	def getRawEntrySortKey(
		key: Optional[Callable[[str], Any]] = None,
	) -> Callable[[Tuple], str]:
		raise NotImplementedError



from os.path import realpath, dirname, join, isdir
import sys

uiDir = ''

if hasattr(sys, 'frozen'):
	rootDir = dirname(sys.executable)
	uiDir = join(rootDir, 'ui')
else:
	uiDir = dirname(realpath(__file__))
	rootDir = dirname(uiDir)

appResDir = join(rootDir, 'res')



# -*- coding: utf-8 -*-
import sys

import os
from os.path import (
	dirname,
	realpath,
)
import re
import gzip
from time import time as now
from collections import Counter
#from typing_extensions import Literal

from pyglossary.text_utils import (
	intToBinStr,
	binStrToInt,
	runDictzip,
)

from formats_common import *

enable = True
format = "Stardict"
description = "StarDict (ifo)"
extensions = [".ifo"]
optionsProp = {
	"dictzip": BoolOption(),
	"sametypesequence": StrOption(
		values=["", "h", "m"],
	),
}
depends = {}
sortOnWrite = ALWAYS
# sortKey also is defined in line 52
supportsAlternates = True

infoKeys = (
	"bookname",
	"author",
	"email",
	"website",
	"description",
	"date",
)


def sortKeyBytes(b_word: bytes) -> Tuple[bytes, bytes]:
	return (
		b_word.lower(),
		b_word,
	)


def sortKey(word: str) -> Tuple[bytes, bytes]:
	return sortKeyBytes(word.encode("utf-8"))


def newlinesToSpace(text: str) -> str:
	return re.sub("[\n\r]+", " ", text)


def newlinesToBr(text: str) -> str:
	return re.sub("\n\r?|\r\n?", "<br>", text)


def verifySameTypeSequence(s: str) -> bool:
	if not s:
		return True
	if not s.isalpha():
		log.error("Invalid sametypesequence option")
		return False
	return True


class Reader(object):
	def __init__(self, glos: GlossaryType):
		self._glos = glos
		self.clear()
		"""
		indexData format
		indexData[i] - i-th record in index file, a tuple (previously a list) of length 3
		indexData[i][0] - b_word (bytes)
		indexData[i][1] - definition block offset in dict file (int)
		indexData[i][2] - definition block size in dict file (int)
		REMOVED:
			indexData[i][3] - list of definitions
			indexData[i][3][j][0] - definition data
			indexData[i][3][j][1] - definition type - "h", "m" or "x"
			indexData[i][4] - list of synonyms (strings)

		synDict:
			a dict { wordIndex -> altList }
		"""

	def close(self) -> None:
		if self._dictFile:
			self._dictFile.close()
		self.clear()

	def clear(self) -> None:
		self._dictFile = None
		self._filename = ""  # base file path, no extension
		self._indexData = []
		self._synDict = {}
		self._sametypesequence = ""
		self._resDir = ""
		self._resFileNames = []
		self._wordCount = None

	def open(self, filename: str) -> None:
		if splitext(filename)[1].lower() == ".ifo":
			self._filename = splitext(filename)[0]
		else:
			self._filename = filename
		self._filename = realpath(self._filename)
		self.readIfoFile()
		sametypesequence = self._glos.getInfo("sametypesequence")
		if not verifySameTypeSequence(sametypesequence):
			return False
		self._indexData = self.readIdxFile()
		self._wordCount = len(self._indexData)
		self._synDict = self.readSynFile()
		self._sametypesequence = sametypesequence
		if isfile(self._filename + ".dict.dz"):
			self._dictFile = gzip.open(self._filename+".dict.dz", mode="rb")
		else:
			self._dictFile = open(self._filename+".dict", mode="rb")
		self._resDir = join(dirname(self._filename), "res")
		if isdir(self._resDir):
			self._resFileNames = os.listdir(self._resDir)
		else:
			self._resDir = ""
			self._resFileNames = []
		# self.readResources()

	def __len__(self) -> int:
		if self._wordCount is None:
			raise RuntimeError(
				"StarDict: len(reader) called while reader is not open"
			)
		return self._wordCount + len(self._resFileNames)

	def readIfoFile(self) -> None:
		"""
		.ifo file is a text file in utf-8 encoding
		"""
		with open(self._filename+".ifo", "r", encoding="utf-8") as ifoFile:
			for line in ifoFile:
				line = line.strip()
				if not line:
					continue
				if line == "StarDict's dict ifo file":
					continue
				key, eq, value = line.partition("=")
				if not (key and value):
					log.warning("Invalid ifo file line: %s", line)
					continue
				self._glos.setInfo(key, value)

	def readIdxFile(self) -> List[Tuple[bytes, int, int]]:
		if isfile(self._filename+".idx.gz"):
			with gzip.open(self._filename+".idx.gz") as idxFile:
				idxBytes = idxFile.read()
		else:
			with open(self._filename+".idx", "rb") as idxFile:
				idxBytes = idxFile.read()

		indexData = []
		pos = 0
		while pos < len(idxBytes):
			beg = pos
			pos = idxBytes.find(b"\x00", beg)
			if pos < 0:
				log.error("Index file is corrupted")
				break
			b_word = idxBytes[beg:pos]
			pos += 1
			if pos + 8 > len(idxBytes):
				log.error("Index file is corrupted")
				break
			offset = binStrToInt(idxBytes[pos:pos+4])
			pos += 4
			size = binStrToInt(idxBytes[pos:pos+4])
			pos += 4
			indexData.append((b_word, offset, size))

		return indexData

	def __iter__(self) -> Iterator[BaseEntry]:
		indexData = self._indexData
		synDict = self._synDict
		sametypesequence = self._sametypesequence
		dictFile = self._dictFile

		if not dictFile:
			log.error("%s is not open, can not iterate", self)
			raise StopIteration

		if not indexData:
			log.warning("indexData is empty")
			raise StopIteration

		for wordIndex, (b_word, defiOffset, defiSize) in enumerate(indexData):
			if not b_word:
				continue

			dictFile.seek(defiOffset)
			if dictFile.tell() != defiOffset:
				log.error("Unable to read definition for word \"%s\"", b_word)
				continue

			b_defiBlock = dictFile.read(defiSize)

			if len(b_defiBlock) != defiSize:
				log.error("Unable to read definition for word \"%s\"", b_word)
				continue

			if sametypesequence:
				defisData = self.parseDefiBlockCompact(
					b_defiBlock,
					sametypesequence,
				)
			else:
				defisData = self.parseDefiBlockGeneral(b_defiBlock)

			if defisData is None:
				log.error("Data file is corrupted. Word \"%s\"", b_word)
				continue

			# defisData is a list of (b_defi, defiFormatCode) tuples

			defis = []
			defiFormats = []
			for b_defi, defiFormatCode in defisData:
				defis.append(b_defi.decode("utf-8"))
				defiFormats.append(
					{
						"m": "m",
						"t": "m",
						"y": "m",
						"g": "h",
						"h": "h",
						"x": "x",
					}.get(chr(defiFormatCode), "")
				)

			# FIXME
			defiFormat = defiFormats[0]
			# defiFormat = Counter(defiFormats).most_common(1)[0][0]

			if not defiFormat:
				log.warning("Definition format %s is not supported", defiFormat)

			word = b_word.decode("utf-8")
			try:
				alts = synDict[wordIndex]
			except KeyError:  # synDict is dict
				pass
			else:
				word = [word] + alts
			if len(defis) == 1:
				defis = defis[0]

			yield self._glos.newEntry(
				word,
				defis,
				defiFormat=defiFormat,
			)

		if isdir(self._resDir):
			for fname in os.listdir(self._resDir):
				fpath = join(self._resDir, fname)
				with open(fpath, "rb") as fromFile:
					yield self._glos.newDataEntry(
						fname,
						fromFile.read(),
					)

	def readSynFile(self) -> Dict[int, List[str]]:
		"""
		return synDict, a dict { wordIndex -> altList }
		"""
		if not isfile(self._filename+".syn"):
			return {}
		with open(self._filename+".syn", "rb") as synFile:
			synBytes = synFile.read()
		synBytesLen = len(synBytes)
		synDict = {}
		pos = 0
		while pos < synBytesLen:
			beg = pos
			pos = synBytes.find(b"\x00", beg)
			if pos < 0:
				log.error("Synonym file is corrupted")
				break
			b_alt = synBytes[beg:pos]  # b_alt is bytes
			pos += 1
			if pos + 4 > len(synBytes):
				log.error("Synonym file is corrupted")
				break
			wordIndex = binStrToInt(synBytes[pos:pos+4])
			pos += 4
			if wordIndex >= self._wordCount:
				log.error(
					"Corrupted synonym file. " +
					"Word \"%s\" references invalid item" % b_alt
				)
				continue

			s_alt = b_alt.decode("utf-8")  # s_alt is str
			try:
				synDict[wordIndex].append(s_alt)
			except KeyError:
				synDict[wordIndex] = [s_alt]

		return synDict

	def parseDefiBlockCompact(self, b_block: bytes, sametypesequence: str) -> List[Tuple[bytes, int]]:
		"""
		Parse definition block when sametypesequence option is specified.

		Return a list of (b_defi, defiFormatCode) tuples
			where b_defi is a bytes instance
			and defiFormatCode is int, so: defiFormat = chr(defiFormatCode)
		"""
		b_sametypesequence = sametypesequence.encode("utf-8")
		assert len(b_sametypesequence) > 0
		res = []
		i = 0
		for t in b_sametypesequence[:-1]:
			if i >= len(b_block):
				return None
			if bytes([t]).islower():
				beg = i
				i = b_block.find(b"\x00", beg)
				if i < 0:
					return None
				res.append((b_block[beg:i], t))
				i += 1
			else:
				assert bytes([t]).isupper()
				if i + 4 > len(b_block):
					return None
				size = binStrToInt(b_block[i:i+4])
				i += 4
				if i + size > len(b_block):
					return None
				res.append((b_block[i:i+size], t))
				i += size

		if i >= len(b_block):
			return None
		t = b_sametypesequence[-1]
		if bytes([t]).islower():
			if 0 in b_block[i:]:
				return None
			res.append((b_block[i:], t))
		else:
			assert bytes([t]).isupper()
			res.append((b_block[i:], t))

		return res

	def parseDefiBlockGeneral(self, b_block: bytes) -> List[Tuple[bytes, int]]:
		"""
		Parse definition block when sametypesequence option is not specified.

		Return a list of (b_defi, defiFormatCode) tuples
			where b_defi is a bytes instance
			and defiFormatCode is int, so: defiFormat = chr(defiFormatCode)
		"""
		res = []
		i = 0
		while i < len(b_block):
			t = b_block[i]
			if not bytes([t]).isalpha():
				return None
			i += 1
			if bytes([t]).islower():
				beg = i
				i = b_block.find(b"\x00", beg)
				if i < 0:
					return None
				res.append((b_block[beg:i], t))
				i += 1
			else:
				assert bytes([t]).isupper()
				if i + 4 > len(b_block):
					return None
				size = binStrToInt(b_block[i:i+4])
				i += 4
				if i + size > len(b_block):
					return None
				res.append((b_block[i:i+size], t))
				i += size
		return res

	# def readResources(self):
	#	if not isdir(self._resDir):
	#		resInfoPath = join(baseDirPath, "res.rifo")
	#		if isfile(resInfoPath):
	#			log.warning(
	#				"StarDict resource database is not supported. Skipping"
	#			)


class Writer(object):
	def __init__(self, glos: GlossaryType):
		self._glos = glos

	def write(
		self,
		filename: str,
		dictzip: bool = True,
		sametypesequence: str = "", # type: Literal["", "h", "m"]
	) -> None:
		fileBasePath = filename
		##
		if splitext(filename)[1].lower() == ".ifo":
			fileBasePath = splitext(filename)[0]
		elif filename.endswith(os.sep):
			if not isdir(filename):
				os.makedirs(filename)
			fileBasePath = join(filename, split(filename[:-1])[-1])
		elif isdir(filename):
			fileBasePath = join(filename, split(filename)[-1])
		##
		if fileBasePath:
			fileBasePath = realpath(fileBasePath)
		self._filename = fileBasePath
		self._resDir = join(dirname(self._filename), "res")


		if sametypesequence:
			log.debug("Using write option sametypesequence=%s" % sametypesequence)
			self.writeCompact(sametypesequence)
#		elif self.glossaryHasAdditionalDefinitions():
#			self.writeGeneral()
		else:
#			defiFormat = self.detectMainDefinitionFormat()
#			if defiFormat == None:
			self.writeGeneral()
#			else:
#				self.writeCompact(defiFormat)

		if dictzip:
			runDictzip(self._filename)

	def writeCompact(self, defiFormat):
		"""
		Build StarDict dictionary with sametypesequence option specified.
		Every item definition consists of a single article.
		All articles have the same format, specified in defiFormat parameter.

		Parameters:
		defiFormat - format of article definition: h - html, m - plain text
		"""
		dictMark = 0
		altIndexList = []  # list of tuples (b"alternate", wordIndex)

		dictFile = open(self._filename+".dict", "wb")
		idxFile = open(self._filename+".idx", "wb")
		indexFileSize = 0

		t0 = now()
		wordCount = 0
		if not isdir(self._resDir):
			os.mkdir(self._resDir)

		entryI = -1
		for entry in self._glos:
			if entry.isData():
				entry.save(self._resDir)
				continue
			entryI += 1

			words = entry.getWords()  # list of strs
			word = words[0]  # str
			defis = entry.getDefis()  # list of strs

			b_dictBlock = b""

			for alt in words[1:]:
				altIndexList.append((alt.encode("utf-8"), entryI))

			b_dictBlock += (defis[0]).encode("utf-8")

			for altDefi in defis[1:]:
				b_dictBlock += b"\x00" + (altDefi).encode("utf-8")

			dictFile.write(b_dictBlock)

			blockLen = len(b_dictBlock)
			b_idxBlock = word.encode("utf-8") + b"\x00" + \
				intToBinStr(dictMark, 4) + \
				intToBinStr(blockLen, 4)
			idxFile.write(b_idxBlock)

			dictMark += blockLen
			indexFileSize += len(b_idxBlock)

			wordCount += 1

		dictFile.close()
		idxFile.close()
		if not os.listdir(self._resDir):
			os.rmdir(self._resDir)
		log.info("Writing dict file took %.2f seconds", now() - t0)
		log.debug("defiFormat = " + pformat(defiFormat))

		self.writeSynFile(altIndexList)
		self.writeIfoFile(wordCount, indexFileSize, len(altIndexList),defiFormat)

	def writeGeneral(self) -> None:
		"""
		Build StarDict dictionary in general case.
		Every item definition may consist of an arbitrary number of articles.
		sametypesequence option is not used.
		"""
		dictMark = 0
		altIndexList = []  # list of tuples (b"alternate", wordIndex)

		dictFile = open(self._filename+".dict", "wb")
		idxFile = open(self._filename+".idx", "wb")
		indexFileSize = 0

		t0 = now()
		wordCount = 0
		defiFormatCounter = Counter()
		if not isdir(self._resDir):
			os.mkdir(self._resDir)

		entryI = -1
		for entry in self._glos:
			if entry.isData():
				entry.save(self._resDir)
				continue
			entryI += 1

			words = entry.getWords()  # list of strs
			word = words[0]  # str
			defis = entry.getDefis()  # list of strs

			entry.detectDefiFormat()  # call no more than once
			defiFormat = entry.getDefiFormat()
			defiFormatCounter[defiFormat] += 1
			if defiFormat not in ("m", "h"):
				defiFormat = "m"

			b_dictBlock = b""

			for alt in words[1:]:
				altIndexList.append((alt.encode("utf-8"), entryI))

			b_dictBlock += (defiFormat + defis[0]).encode("utf-8") + b"\x00"

			for altDefi in defis[1:]:
				b_dictBlock += (defiFormat + altDefi).encode("utf-8") + b"\x00"

			dictFile.write(b_dictBlock)

			blockLen = len(b_dictBlock)
			b_idxBlock = word.encode("utf-8") + b"\x00" + \
				intToBinStr(dictMark, 4) + \
				intToBinStr(blockLen, 4)
			idxFile.write(b_idxBlock)

			dictMark += blockLen
			indexFileSize += len(b_idxBlock)

			wordCount += 1

		dictFile.close()
		idxFile.close()
		if not os.listdir(self._resDir):
			os.rmdir(self._resDir)
		log.info("Writing dict file took %.2f seconds", now() - t0)
		log.debug("defiFormatsCount = " + pformat(defiFormatCounter.most_common()))

		self.writeSynFile(altIndexList)
		self.writeIfoFile(wordCount, indexFileSize, len(altIndexList))

	def writeSynFile(self, altIndexList: List[Tuple[bytes, int]]) -> None:
		"""
		Build .syn file
		"""
		if not altIndexList:
			return

		log.info("Sorting %s synonyms...", len(altIndexList))
		t0 = now()

		altIndexList.sort(
			key=lambda x: sortKeyBytes(x[0])
		)
		# 28 seconds with old sort key (converted from custom cmp)
		# 0.63 seconds with my new sort key
		# 0.20 seconds without key function (default sort)

		log.info(
			"Sorting %s synonyms took %.2f seconds",
			len(altIndexList),
			now() - t0,
		)
		log.info("Writing %s synonyms...", len(altIndexList))
		t0 = now()
		with open(self._filename+".syn", "wb") as synFile:
			synFile.write(b"".join([
				b_alt + b"\x00" + intToBinStr(wordIndex, 4)
				for b_alt, wordIndex in altIndexList
			]))
		log.info(
			"Writing %s synonyms took %.2f seconds",
			len(altIndexList),
			now() - t0,
		)

	def writeIfoFile(
		self,
		wordCount: int,
		indexFileSize: int,
		synwordcount: int,
		sametypesequence: str = "", # type: Literal["", "h", "m"]
	) -> None:
		"""
		Build .ifo file
		"""
		ifoStr = "StarDict's dict ifo file\n" \
			+ "version=3.0.0\n" \
			+ "bookname=%s\n" % newlinesToSpace(self._glos.getInfo("name")) \
			+ "wordcount=%s\n" % wordCount \
			+ "idxfilesize=%s\n" % indexFileSize
		if sametypesequence:
			ifoStr += "sametypesequence=%s\n" % sametypesequence
		if synwordcount > 0:
			ifoStr += "synwordcount=%s\n" % synwordcount
		for key in infoKeys:
			if key in (
				"bookname",
				"wordcount",
				"idxfilesize",
				"sametypesequence",
			):
				continue
			value = self._glos.getInfo(key)
			if value == "":
				continue
			if key == "description":
				value = newlinesToBr(value)
			else:
				value = newlinesToSpace(value)

			ifoStr += "%s=%s\n" % (key, value)

		with open(self._filename+".ifo", "w", encoding="utf-8") as ifoFile:
			ifoFile.write(ifoStr)

	def glossaryHasAdditionalDefinitions(self) -> bool:
		"""
		Search for additional definitions in the glossary.
		We need to know if the glossary contains additional definitions
		to make the decision on the format of the StarDict dictionary.
		"""
		for entry in self._glos:
			if len(entry.getDefis()) > 1:
				return True
		return False

	def detectMainDefinitionFormat(self) -> Optional[str]:
		"""
		Scan main definitions of the glossary.
		Return format common to all definitions: "h" or "m"

		If definitions has different formats return None.
		"""
		self._glos.setDefaultDefiFormat("m")
		formatsCount = self._glos.getMostUsedDefiFormats()
		if not formatsCount:
			return None
		if len(formatsCount) > 1:  # FIXME
			return None

		return formatsCount[0]




# -*- coding: utf-8 -*-
##
## Copyright © 2012 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
## This file is part of PyGlossary project, https://github.com/ilius/pyglossary
##
## This program is a free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3, or (at your option)
## any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
## If not, see <http://www.gnu.org/licenses/gpl.txt>.

from os.path import join, isfile

from pyglossary.core import (
	rootConfJsonFile,
	confJsonFile,
	rootDir,
	dataDir,
	appResDir,
)
from pyglossary.glossary import *
from pyglossary.json_utils import jsonToData

def fread(path):
	with open(path, encoding="utf-8") as fp:
		return fp.read()

logo = join(appResDir, "pyglossary.png")
aboutText = fread(join(dataDir, "about"))
licenseText = fread(join(dataDir, "license-dialog"))
authors = fread(join(dataDir, "AUTHORS")).split("\n")


class UIBase(object):
	prefKeys = (
		"noProgressBar",## command line
		"ui_autoSetFormat",
		"ui_autoSetOutputFileName",
		"lower",
		"utf8Check",
		"remove_html",
		"remove_html_all",
		"enable_alts",
		## Reverse Options:
		"reverse_matchWord",
		"reverse_showRel",
		"reverse_saveStep",
		"reverse_minRel",
		"reverse_maxNum",
		"reverse_includeDefs",
	)
	def pref_load(self, **options):
		data = jsonToData(fread(rootConfJsonFile))
		if isfile(confJsonFile):
			try:
				userData = jsonToData(fread(confJsonFile))
			except Exception:
				log.exception("error while loading user config file \"%s\"", confJsonFile)
			else:
				data.update(userData)

		for key in self.prefKeys:
			try:
				self.pref[key] = data.pop(key)
			except KeyError:
				pass
		for key, value in data.items():
			log.warning("unkown config key \"%s\"", key)

		for key, value in options.items():
			if key in self.prefKeys:
				self.pref[key] = value
		
		return True

	def progressEnd(self):
		self.progress(1.0)




